"""Python wrappers around TensorFlow ops.

This file is MACHINE GENERATED! Do not edit.
"""

import collections as _collections
import six as _six

from tensorflow.python import pywrap_tensorflow as _pywrap_tensorflow
from tensorflow.python.eager import context as _context
from tensorflow.python.eager import core as _core
from tensorflow.python.eager import execute as _execute
from tensorflow.python.framework import dtypes as _dtypes
from tensorflow.python.framework import errors as _errors
from tensorflow.python.framework import tensor_shape as _tensor_shape

from tensorflow.core.framework import op_def_pb2 as _op_def_pb2
# Needed to trigger the call to _set_call_cpp_shape_fn.
from tensorflow.python.framework import common_shapes as _common_shapes
from tensorflow.python.framework import op_def_registry as _op_def_registry
from tensorflow.python.framework import ops as _ops
from tensorflow.python.framework import op_def_library as _op_def_library
from tensorflow.python.util.deprecation import deprecated_endpoints
from tensorflow.python.util import dispatch as _dispatch
from tensorflow.python.util.tf_export import tf_export
from tensorflow.python.util.tf_export import kwarg_only as _kwarg_only
from tensorflow.tools.docs import doc_controls as _doc_controls


@_dispatch.add_dispatch_list
@tf_export('rpc')
def rpc(address, method, request, protocol="", fail_fast=True, timeout_in_ms=0, name=None):
  r"""TODO: add doc.

  Args:
    address: A `Tensor` of type `string`.
    method: A `Tensor` of type `string`.
    request: A `Tensor` of type `string`.
    protocol: An optional `string`. Defaults to `""`.
    fail_fast: An optional `bool`. Defaults to `True`.
    timeout_in_ms: An optional `int`. Defaults to `0`.
    name: A name for the operation (optional).

  Returns:
    A `Tensor` of type `string`.
  """
  _ctx = _context._context or _context.context()
  if _ctx is not None and _ctx._thread_local_data.is_eager:
    try:
      _result = _pywrap_tensorflow.TFE_Py_FastPathExecute(
        _ctx._context_handle, _ctx._thread_local_data.device_name, "Rpc",
        name, _ctx.post_execution_callbacks, address, method, request,
        "protocol", protocol, "fail_fast", fail_fast, "timeout_in_ms",
        timeout_in_ms)
      return _result
    except _core._FallbackException:
      try:
        return rpc_eager_fallback(
            address, method, request, protocol=protocol, fail_fast=fail_fast,
            timeout_in_ms=timeout_in_ms, name=name, ctx=_ctx)
      except _core._SymbolicException:
        pass  # Add nodes to the TensorFlow graph.
      except (TypeError, ValueError):
        result = _dispatch.dispatch(
              rpc, address=address, method=method, request=request,
                   protocol=protocol, fail_fast=fail_fast,
                   timeout_in_ms=timeout_in_ms, name=name)
        if result is not _dispatch.OpDispatcher.NOT_SUPPORTED:
          return result
        raise
    except _core._NotOkStatusException as e:
      if name is not None:
        message = e.message + " name: " + name
      else:
        message = e.message
      _six.raise_from(_core._status_to_exception(e.code, message), None)
  # Add nodes to the TensorFlow graph.
  if protocol is None:
    protocol = ""
  protocol = _execute.make_str(protocol, "protocol")
  if fail_fast is None:
    fail_fast = True
  fail_fast = _execute.make_bool(fail_fast, "fail_fast")
  if timeout_in_ms is None:
    timeout_in_ms = 0
  timeout_in_ms = _execute.make_int(timeout_in_ms, "timeout_in_ms")
  try:
    _, _, _op = _op_def_lib._apply_op_helper(
        "Rpc", address=address, method=method, request=request,
               protocol=protocol, fail_fast=fail_fast,
               timeout_in_ms=timeout_in_ms, name=name)
  except (TypeError, ValueError):
    result = _dispatch.dispatch(
          rpc, address=address, method=method, request=request,
               protocol=protocol, fail_fast=fail_fast,
               timeout_in_ms=timeout_in_ms, name=name)
    if result is not _dispatch.OpDispatcher.NOT_SUPPORTED:
      return result
    raise
  _result = _op.outputs[:]
  _inputs_flat = _op.inputs
  _attrs = ("protocol", _op.get_attr("protocol"), "fail_fast",
            _op.get_attr("fail_fast"), "timeout_in_ms",
            _op.get_attr("timeout_in_ms"))
  _execute.record_gradient(
      "Rpc", _inputs_flat, _attrs, _result, name)
  _result, = _result
  return _result

def Rpc(address, method, request, protocol="", fail_fast=True, timeout_in_ms=0, name=None):
  return rpc(address=address, method=method, request=request, protocol=protocol, fail_fast=fail_fast, timeout_in_ms=timeout_in_ms, name=name)
Rpc.__doc__ = rpc.__doc__
Rpc = _doc_controls.do_not_generate_docs(_kwarg_only(Rpc))
tf_export("raw_ops.Rpc")(Rpc)


def rpc_eager_fallback(address, method, request, protocol="", fail_fast=True, timeout_in_ms=0, name=None, ctx=None):
  r"""This is the slowpath function for Eager mode.
  This is for function rpc
  """
  _ctx = ctx if ctx else _context.context()
  if protocol is None:
    protocol = ""
  protocol = _execute.make_str(protocol, "protocol")
  if fail_fast is None:
    fail_fast = True
  fail_fast = _execute.make_bool(fail_fast, "fail_fast")
  if timeout_in_ms is None:
    timeout_in_ms = 0
  timeout_in_ms = _execute.make_int(timeout_in_ms, "timeout_in_ms")
  address = _ops.convert_to_tensor(address, _dtypes.string)
  method = _ops.convert_to_tensor(method, _dtypes.string)
  request = _ops.convert_to_tensor(request, _dtypes.string)
  _inputs_flat = [address, method, request]
  _attrs = ("protocol", protocol, "fail_fast", fail_fast, "timeout_in_ms",
  timeout_in_ms)
  _result = _execute.execute(b"Rpc", 1, inputs=_inputs_flat, attrs=_attrs,
                             ctx=_ctx, name=name)
  _execute.record_gradient(
      "Rpc", _inputs_flat, _attrs, _result, name)
  _result, = _result
  return _result

_ops.RegisterShape("Rpc")(None)


_try_rpc_outputs = ["response", "status_code", "status_message"]
_TryRpcOutput = _collections.namedtuple(
    "TryRpc", _try_rpc_outputs)


@_dispatch.add_dispatch_list
@tf_export('try_rpc')
def try_rpc(address, method, request, protocol="", fail_fast=True, timeout_in_ms=0, name=None):
  r"""TODO: add doc.

  Args:
    address: A `Tensor` of type `string`.
    method: A `Tensor` of type `string`.
    request: A `Tensor` of type `string`.
    protocol: An optional `string`. Defaults to `""`.
    fail_fast: An optional `bool`. Defaults to `True`.
    timeout_in_ms: An optional `int`. Defaults to `0`.
    name: A name for the operation (optional).

  Returns:
    A tuple of `Tensor` objects (response, status_code, status_message).

    response: A `Tensor` of type `string`.
    status_code: A `Tensor` of type `int32`.
    status_message: A `Tensor` of type `string`.
  """
  _ctx = _context._context or _context.context()
  if _ctx is not None and _ctx._thread_local_data.is_eager:
    try:
      _result = _pywrap_tensorflow.TFE_Py_FastPathExecute(
        _ctx._context_handle, _ctx._thread_local_data.device_name, "TryRpc",
        name, _ctx.post_execution_callbacks, address, method, request,
        "protocol", protocol, "fail_fast", fail_fast, "timeout_in_ms",
        timeout_in_ms)
      _result = _TryRpcOutput._make(_result)
      return _result
    except _core._FallbackException:
      try:
        return try_rpc_eager_fallback(
            address, method, request, protocol=protocol, fail_fast=fail_fast,
            timeout_in_ms=timeout_in_ms, name=name, ctx=_ctx)
      except _core._SymbolicException:
        pass  # Add nodes to the TensorFlow graph.
      except (TypeError, ValueError):
        result = _dispatch.dispatch(
              try_rpc, address=address, method=method, request=request,
                       protocol=protocol, fail_fast=fail_fast,
                       timeout_in_ms=timeout_in_ms, name=name)
        if result is not _dispatch.OpDispatcher.NOT_SUPPORTED:
          return result
        raise
    except _core._NotOkStatusException as e:
      if name is not None:
        message = e.message + " name: " + name
      else:
        message = e.message
      _six.raise_from(_core._status_to_exception(e.code, message), None)
  # Add nodes to the TensorFlow graph.
  if protocol is None:
    protocol = ""
  protocol = _execute.make_str(protocol, "protocol")
  if fail_fast is None:
    fail_fast = True
  fail_fast = _execute.make_bool(fail_fast, "fail_fast")
  if timeout_in_ms is None:
    timeout_in_ms = 0
  timeout_in_ms = _execute.make_int(timeout_in_ms, "timeout_in_ms")
  try:
    _, _, _op = _op_def_lib._apply_op_helper(
        "TryRpc", address=address, method=method, request=request,
                  protocol=protocol, fail_fast=fail_fast,
                  timeout_in_ms=timeout_in_ms, name=name)
  except (TypeError, ValueError):
    result = _dispatch.dispatch(
          try_rpc, address=address, method=method, request=request,
                   protocol=protocol, fail_fast=fail_fast,
                   timeout_in_ms=timeout_in_ms, name=name)
    if result is not _dispatch.OpDispatcher.NOT_SUPPORTED:
      return result
    raise
  _result = _op.outputs[:]
  _inputs_flat = _op.inputs
  _attrs = ("protocol", _op.get_attr("protocol"), "fail_fast",
            _op.get_attr("fail_fast"), "timeout_in_ms",
            _op.get_attr("timeout_in_ms"))
  _execute.record_gradient(
      "TryRpc", _inputs_flat, _attrs, _result, name)
  _result = _TryRpcOutput._make(_result)
  return _result

def TryRpc(address, method, request, protocol="", fail_fast=True, timeout_in_ms=0, name=None):
  return try_rpc(address=address, method=method, request=request, protocol=protocol, fail_fast=fail_fast, timeout_in_ms=timeout_in_ms, name=name)
TryRpc.__doc__ = try_rpc.__doc__
TryRpc = _doc_controls.do_not_generate_docs(_kwarg_only(TryRpc))
tf_export("raw_ops.TryRpc")(TryRpc)


def try_rpc_eager_fallback(address, method, request, protocol="", fail_fast=True, timeout_in_ms=0, name=None, ctx=None):
  r"""This is the slowpath function for Eager mode.
  This is for function try_rpc
  """
  _ctx = ctx if ctx else _context.context()
  if protocol is None:
    protocol = ""
  protocol = _execute.make_str(protocol, "protocol")
  if fail_fast is None:
    fail_fast = True
  fail_fast = _execute.make_bool(fail_fast, "fail_fast")
  if timeout_in_ms is None:
    timeout_in_ms = 0
  timeout_in_ms = _execute.make_int(timeout_in_ms, "timeout_in_ms")
  address = _ops.convert_to_tensor(address, _dtypes.string)
  method = _ops.convert_to_tensor(method, _dtypes.string)
  request = _ops.convert_to_tensor(request, _dtypes.string)
  _inputs_flat = [address, method, request]
  _attrs = ("protocol", protocol, "fail_fast", fail_fast, "timeout_in_ms",
  timeout_in_ms)
  _result = _execute.execute(b"TryRpc", 3, inputs=_inputs_flat, attrs=_attrs,
                             ctx=_ctx, name=name)
  _execute.record_gradient(
      "TryRpc", _inputs_flat, _attrs, _result, name)
  _result = _TryRpcOutput._make(_result)
  return _result

_ops.RegisterShape("TryRpc")(None)

def _InitOpDefLibrary(op_list_proto_bytes):
  op_list = _op_def_pb2.OpList()
  op_list.ParseFromString(op_list_proto_bytes)
  _op_def_registry.register_op_list(op_list)
  op_def_lib = _op_def_library.OpDefLibrary()
  op_def_lib.add_op_list(op_list)
  return op_def_lib
# op {
#   name: "Rpc"
#   input_arg {
#     name: "address"
#     type: DT_STRING
#   }
#   input_arg {
#     name: "method"
#     type: DT_STRING
#   }
#   input_arg {
#     name: "request"
#     type: DT_STRING
#   }
#   output_arg {
#     name: "response"
#     type: DT_STRING
#   }
#   attr {
#     name: "protocol"
#     type: "string"
#     default_value {
#       s: ""
#     }
#   }
#   attr {
#     name: "fail_fast"
#     type: "bool"
#     default_value {
#       b: true
#     }
#   }
#   attr {
#     name: "timeout_in_ms"
#     type: "int"
#     default_value {
#       i: 0
#     }
#   }
#   is_stateful: true
# }
# op {
#   name: "TryRpc"
#   input_arg {
#     name: "address"
#     type: DT_STRING
#   }
#   input_arg {
#     name: "method"
#     type: DT_STRING
#   }
#   input_arg {
#     name: "request"
#     type: DT_STRING
#   }
#   output_arg {
#     name: "response"
#     type: DT_STRING
#   }
#   output_arg {
#     name: "status_code"
#     type: DT_INT32
#   }
#   output_arg {
#     name: "status_message"
#     type: DT_STRING
#   }
#   attr {
#     name: "protocol"
#     type: "string"
#     default_value {
#       s: ""
#     }
#   }
#   attr {
#     name: "fail_fast"
#     type: "bool"
#     default_value {
#       b: true
#     }
#   }
#   attr {
#     name: "timeout_in_ms"
#     type: "int"
#     default_value {
#       i: 0
#     }
#   }
#   is_stateful: true
# }
_op_def_lib = _InitOpDefLibrary(b"\n\205\001\n\003Rpc\022\013\n\007address\030\007\022\n\n\006method\030\007\022\013\n\007request\030\007\032\014\n\010response\030\007\"\026\n\010protocol\022\006string\032\002\022\000\"\025\n\tfail_fast\022\004bool\032\002(\001\"\030\n\rtimeout_in_ms\022\003int\032\002\030\000\210\001\001\n\255\001\n\006TryRpc\022\013\n\007address\030\007\022\n\n\006method\030\007\022\013\n\007request\030\007\032\014\n\010response\030\007\032\017\n\013status_code\030\003\032\022\n\016status_message\030\007\"\026\n\010protocol\022\006string\032\002\022\000\"\025\n\tfail_fast\022\004bool\032\002(\001\"\030\n\rtimeout_in_ms\022\003int\032\002\030\000\210\001\001")



# Copyright 2016 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Run Config (deprecated, use tf.estimator.RunConfig instead).

This module and all its submodules are deprecated. See
[contrib/learn/README.md](https://www.tensorflow.org/code/tensorflow/contrib/learn/README.md)
for migration instructions.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import json
import os

import six

from tensorflow.contrib.framework.python.framework import experimental
from tensorflow.core.protobuf import config_pb2
from tensorflow.python.estimator import run_config as core_run_config
from tensorflow.python.platform import tf_logging as logging
from tensorflow.python.training import server_lib
from tensorflow.python.util.deprecation import deprecated


# A list of the property names in RunConfig user allows to change. They will
# not affect the execution framework, so when execution framework checks the
# `uid` of the RunConfig, it should be ignored.
_DEFAULT_UID_WHITE_LIST = [
    'tf_random_seed',
    'save_summary_steps',
    'save_checkpoints_steps',
    'save_checkpoints_secs',
    'session_config',
    'keep_checkpoint_max',
    'keep_checkpoint_every_n_hours',
    'log_step_count_steps',
]


class Environment(object):
  """DEPRECATED CLASS."""
  # For running general distributed training.
  CLOUD = 'cloud'
  # For running Google-internal distributed training.
  GOOGLE = 'google'
  # For running on local desktop.
  LOCAL = 'local'


class TaskType(object):
  """DEPRECATED CLASS."""
  MASTER = 'master'
  PS = 'ps'
  WORKER = 'worker'


class ClusterConfig(object):
  """This class specifies the configurations for a distributed run.

  THIS CLASS IS DEPRECATED. Use tf.estimator.RunConfig instead.

  If you're using an `Estimator`, you should probably use the subclass
  RunConfig instead.
  """

  def __init__(self, master=None, evaluation_master=None):
    """Constructor.

    Sets the properties `cluster_spec`, `is_chief`, `master` (if `None` in the
    args), `num_ps_replicas`, `task_id`, and `task_type` based on the
    `TF_CONFIG` environment variable, if the pertinent information is
    present. The `TF_CONFIG` environment variable is a JSON object with
    attributes: `cluster`, `environment`, and `task`.

    `cluster` is a JSON serialized version of `ClusterSpec`'s Python dict from
    `server_lib.py`, mapping task types (usually one of the TaskType enums) to a
    list of task addresses.

    `environment` specifies the runtime environment for the job (usually one of
    the `Environment` enums). Defaults to `LOCAL`.

    `task` has two attributes: `type` and `index`, where `type` can be any of
    the task types in `cluster`. When `TF_CONFIG` contains said information, the
    following properties are set on this class:

    * `task_type` is set to `TF_CONFIG['task']['type']`. Defaults to `None`.
    * `task_id` is set to `TF_CONFIG['task']['index']`. Defaults to 0.
    * `cluster_spec` is parsed from `TF_CONFIG['cluster']`. Defaults to {}.
    * `master` is determined by looking up `task_type` and `task_id` in the
      `cluster_spec`. Defaults to ''.
    * `num_ps_replicas` is set by counting the number of nodes listed
      in the `ps` attribute of `cluster_spec`. Defaults to 0.
    * `num_worker_replicas` is set by counting the number of nodes listed
      in the `worker` attribute of `cluster_spec`. Defaults to 0.
    * `is_chief` is deteremined based on `task_type`, `type_id`, and
      `environment`.

    Example:
    ```
      cluster = {'ps': ['host1:2222', 'host2:2222'],
                 'worker': ['host3:2222', 'host4:2222', 'host5:2222']}
      os.environ['TF_CONFIG'] = json.dumps(
          {'cluster': cluster,
           'task': {'type': 'worker', 'index': 1}})
      config = ClusterConfig()
      assert config.master == 'host4:2222'
      assert config.task_id == 1
      assert config.num_ps_replicas == 2
      assert config.num_worker_replicas == 3
      assert config.cluster_spec == server_lib.ClusterSpec(cluster)
      assert config.task_type == 'worker'
      assert not config.is_chief
    ```

    Args:
      master: TensorFlow master. Defaults to empty string for local.
      evaluation_master: The master on which to perform evaluation.
    """
    # If not explicitly specified in the constructor and the TF_CONFIG
    # environment variable is present, load cluster_spec from TF_CONFIG.
    config = json.loads(os.environ.get('TF_CONFIG') or '{}')

    # Set task_type and task_id if the TF_CONFIG environment variable is
    # present.  Otherwise, use the respective default (None / 0).
    task_env = config.get('task', {})
    self._task_type = task_env.get('type', None)
    self._task_id = self.get_task_id()

    self._cluster_spec = server_lib.ClusterSpec(config.get('cluster', {}))
    self._master = (master if master is not None else
                    _get_master(self._cluster_spec, self._task_type,
                                self._task_id) or '')
    self._num_ps_replicas = _count_ps(self._cluster_spec) or 0
    self._num_worker_replicas = _count_worker(self._cluster_spec) or 0

    # Set is_chief.
    self._environment = config.get('environment', Environment.LOCAL)
    self._is_chief = None
    if self._task_type is None:
      self._is_chief = (self._task_id == 0)
    elif self._environment == Environment.CLOUD:
      # When the TF_CONFIG environment variable is set, we can set the
      # default of is_chief to 0 when task_type is "master" and task_id is 0.
      self._is_chief = (self._task_type == TaskType.MASTER and
                        self._task_id == 0)
    else:
      # Legacy behavior is that is_chief is None if task_id == 0.
      self._is_chief = (self._task_type == TaskType.WORKER and
                        self._task_id == 0)

    self._evaluation_master = evaluation_master or ''

  @property
  def cluster_spec(self):
    return self._cluster_spec

  @property
  def environment(self):
    return self._environment

  @property
  def evaluation_master(self):
    return self._evaluation_master

  @property
  def is_chief(self):
    return self._is_chief

  @property
  def master(self):
    return self._master

  @property
  def num_ps_replicas(self):
    return self._num_ps_replicas

  @property
  def num_worker_replicas(self):
    return self._num_worker_replicas

  @property
  def task_id(self):
    return self._task_id

  @property
  def task_type(self):
    return self._task_type

  @staticmethod
  def get_task_id():
    """Returns task index from `TF_CONFIG` environmental variable.

    If you have a ClusterConfig instance, you can just access its task_id
    property instead of calling this function and re-parsing the environmental
    variable.

    Returns:
      `TF_CONFIG['task']['index']`. Defaults to 0.
    """
    config = json.loads(os.environ.get('TF_CONFIG') or '{}')
    task_env = config.get('task', {})
    task_index = task_env.get('index')
    return int(task_index) if task_index else 0


class RunConfig(ClusterConfig, core_run_config.RunConfig):
  """This class specifies the configurations for an `Estimator` run.

  This class is a deprecated implementation of `tf.estimator.RunConfig`
  interface.
  """
  _USE_DEFAULT = 0

  @deprecated(None, 'When switching to tf.estimator.Estimator, use'
              ' tf.estimator.RunConfig instead.')
  def __init__(self,
               master=None,
               num_cores=0,
               log_device_placement=False,
               gpu_memory_fraction=1,
               tf_random_seed=None,
               save_summary_steps=100,
               save_checkpoints_secs=_USE_DEFAULT,
               save_checkpoints_steps=None,
               keep_checkpoint_max=5,
               keep_checkpoint_every_n_hours=10000,
               log_step_count_steps=100,
               protocol=None,
               evaluation_master='',
               model_dir=None,
               session_config=None,
               session_creation_timeout_secs=7200):
    """Constructor.

    The superclass `ClusterConfig` may set properties like `cluster_spec`,
    `is_chief`, `master` (if `None` in the args), `num_ps_replicas`, `task_id`,
    and `task_type` based on the `TF_CONFIG` environment variable. See
    `ClusterConfig` for more details.

    N.B.: If `save_checkpoints_steps` or `save_checkpoints_secs` is set,
    `keep_checkpoint_max` might need to be adjusted accordingly, especially in
    distributed training. For example, setting `save_checkpoints_secs` as 60
    without adjusting `keep_checkpoint_max` (defaults to 5) leads to situation
    that checkpoint would be garbage collected after 5 minutes. In distributed
    training, the evaluation job starts asynchronously and might fail to load or
    find the checkpoint due to race condition.

    Args:
      master: TensorFlow master. Defaults to empty string for local.
      num_cores: Number of cores to be used. If 0, the system picks an
        appropriate number (default: 0).
      log_device_placement: Log the op placement to devices (default: False).
      gpu_memory_fraction: Fraction of GPU memory used by the process on
        each GPU uniformly on the same machine.
      tf_random_seed: Random seed for TensorFlow initializers.
        Setting this value allows consistency between reruns.
      save_summary_steps: Save summaries every this many steps.
      save_checkpoints_secs: Save checkpoints every this many seconds. Can not
          be specified with `save_checkpoints_steps`.
      save_checkpoints_steps: Save checkpoints every this many steps. Can not be
          specified with `save_checkpoints_secs`.
      keep_checkpoint_max: The maximum number of recent checkpoint files to
        keep. As new files are created, older files are deleted. If None or 0,
        all checkpoint files are kept. Defaults to 5 (that is, the 5 most recent
        checkpoint files are kept.)
      keep_checkpoint_every_n_hours: Number of hours between each checkpoint
        to be saved. The default value of 10,000 hours effectively disables
        the feature.
      log_step_count_steps: The frequency, in number of global steps, that the
        global step/sec will be logged during training.
      protocol: An optional argument which specifies the protocol used when
        starting server. None means default to grpc.
      evaluation_master: the master on which to perform evaluation.
      model_dir: directory where model parameters, graph etc are saved. If
        `None`, will use `model_dir` property in `TF_CONFIG` environment
        variable. If both are set, must have same value. If both are `None`, see
        `Estimator` about where the model will be saved.
      session_config: a ConfigProto used to set session parameters, or None.
        Note - using this argument, it is easy to provide settings which break
        otherwise perfectly good models. Use with care.
      session_creation_timeout_secs: Max time workers should wait for a session
        to become available (on initialization or when recovering a session)
        with MonitoredTrainingSession. Defaults to 7200 seconds, but users may
        want to set a lower value to detect problems with variable / session
        (re)-initialization more quickly.
    """
    # Neither parent class calls super().__init__(), so here we have to
    # manually call their __init__() methods.
    ClusterConfig.__init__(
        self, master=master, evaluation_master=evaluation_master)
    # For too long this code didn't call:
    #   core_run_config.RunConfig.__init__(self)
    # so instead of breaking compatibility with that assumption, we
    # just manually initialize this field:
    self._train_distribute = None
    self._eval_distribute = None
    self._experimental_max_worker_delay_secs = None
    self._device_fn = None

    gpu_options = config_pb2.GPUOptions(
        per_process_gpu_memory_fraction=gpu_memory_fraction)
    self._tf_config = config_pb2.ConfigProto(
        log_device_placement=log_device_placement,
        inter_op_parallelism_threads=num_cores,
        intra_op_parallelism_threads=num_cores,
        gpu_options=gpu_options)

    self._tf_random_seed = tf_random_seed
    self._save_summary_steps = save_summary_steps
    self._save_checkpoints_secs = save_checkpoints_secs
    self._log_step_count_steps = log_step_count_steps
    self._protocol = protocol
    self._session_config = session_config
    if save_checkpoints_secs == RunConfig._USE_DEFAULT:
      if save_checkpoints_steps is None:
        self._save_checkpoints_secs = 600
      else:
        self._save_checkpoints_secs = None
    self._save_checkpoints_steps = save_checkpoints_steps

    # TODO(weiho): Remove these after ModelFn refactoring, when users can
    # create Scaffold and Saver in their model_fn to set these.
    self._keep_checkpoint_max = keep_checkpoint_max
    self._keep_checkpoint_every_n_hours = keep_checkpoint_every_n_hours
    self._model_dir = _get_model_dir(model_dir)
    self._session_creation_timeout_secs = session_creation_timeout_secs

  @experimental
  def uid(self, whitelist=None):
    """Generates a 'Unique Identifier' based on all internal fields.

    Caller should use the uid string to check `RunConfig` instance integrity
    in one session use, but should not rely on the implementation details, which
    is subject to change.

    Args:
      whitelist: A list of the string names of the properties uid should not
        include. If `None`, defaults to `_DEFAULT_UID_WHITE_LIST`, which
        includes most properties user allowes to change.

    Returns:
      A uid string.
    """
    if whitelist is None:
      whitelist = _DEFAULT_UID_WHITE_LIST

    state = {k: v for k, v in self.__dict__.items() if not k.startswith('__')}
    # Pop out the keys in whitelist.
    for k in whitelist:
      state.pop('_' + k, None)

    ordered_state = collections.OrderedDict(
        sorted(state.items(), key=lambda t: t[0]))
    # For class instance without __repr__, some special cares are required.
    # Otherwise, the object address will be used.
    if '_cluster_spec' in ordered_state:
      ordered_state['_cluster_spec'] = collections.OrderedDict(
          sorted(ordered_state['_cluster_spec'].as_dict().items(),
                 key=lambda t: t[0]))
    return ', '.join(
        '%s=%r' % (k, v) for (k, v) in six.iteritems(ordered_state))

  @property
  def model_dir(self):
    return self._model_dir

  @property
  def tf_config(self):
    return self._tf_config

  @property
  def tf_random_seed(self):
    return self._tf_random_seed

  @property
  def save_summary_steps(self):
    return self._save_summary_steps

  @property
  def save_checkpoints_secs(self):
    return self._save_checkpoints_secs

  @property
  def save_checkpoints_steps(self):
    return self._save_checkpoints_steps

  @property
  def session_config(self):
    return self._session_config

  @property
  def keep_checkpoint_max(self):
    return self._keep_checkpoint_max

  @property
  def keep_checkpoint_every_n_hours(self):
    return self._keep_checkpoint_every_n_hours

  @property
  def log_step_count_steps(self):
    return self._log_step_count_steps

  @property
  def session_creation_timeout_secs(self):
    return self._session_creation_timeout_secs


def _count_ps(cluster_spec):
  """Counts the number of parameter servers in cluster_spec."""
  return len(cluster_spec.as_dict().get('ps', [])) if cluster_spec else 0


def _count_worker(cluster_spec):
  """Counts the number of workers in cluster_spec.

  Workers with TaskType.WORKER and TaskType.MASTER are included in the return
  value.

  Args:
    cluster_spec: a ClusterSpec instance that describes current deployment.

  Returns:
    The total number of eligible workers.

    If 'cluster_spec' was None, then 0 is returned.
  """
  return (len(cluster_spec.as_dict().get('worker', [])) +
          len(cluster_spec.as_dict().get('master', []))) if cluster_spec else 0


def _get_master(cluster_spec, task_type, task_id):
  """Returns the appropriate string for the TensorFlow master."""
  if not cluster_spec:
    return ''

  # If there is only one node in the cluster, do things locally.
  jobs = cluster_spec.jobs
  if len(jobs) == 1 and len(cluster_spec.job_tasks(jobs[0])) == 1:
    return ''

  # Lookup the master in cluster_spec using task_type and task_id,
  # if possible.
  if task_type:
    if task_type not in jobs:
      raise ValueError(
          '%s is not a valid task_type in the cluster_spec:\n'
          '%s\n\n'
          'Note that these values may be coming from the TF_CONFIG environment '
          'variable.' % (task_type, cluster_spec))
    addresses = cluster_spec.job_tasks(task_type)
    if task_id >= len(addresses) or task_id < 0:
      raise ValueError(
          '%d is not a valid task_id for task_type %s in the '
          'cluster_spec:\n'
          '%s\n\n'
          'Note that these value may be coming from the TF_CONFIG environment '
          'variable.' % (task_id, task_type, cluster_spec))
    return 'grpc://' + addresses[task_id]

  # For backwards compatibility, we return empty string if task_type was
  # not set (task_type did not previously exist).
  return ''


def _get_model_dir(model_dir):
  """Returns `model_dir` based user provided `model_dir` or `TF_CONFIG`."""

  model_dir_in_tf_config = json.loads(
      os.environ.get('TF_CONFIG') or '{}').get('model_dir', None)
  if model_dir_in_tf_config is not None:
    if model_dir is not None and model_dir_in_tf_config != model_dir:
      raise ValueError(
          '`model_dir` provided in RunConfig construct, if set, '
          'must have the same value as the model_dir in TF_CONFIG. '
          'model_dir: {}\nTF_CONFIG["model_dir"]: {}.\n'.format(
              model_dir, model_dir_in_tf_config))

    logging.info('Using model_dir in TF_CONFIG: %s', model_dir_in_tf_config)

  return model_dir or model_dir_in_tf_config



"""Python wrappers around TensorFlow ops.

This file is MACHINE GENERATED! Do not edit.
"""

import collections as _collections
import six as _six

from tensorflow.python import pywrap_tensorflow as _pywrap_tensorflow
from tensorflow.python.eager import context as _context
from tensorflow.python.eager import core as _core
from tensorflow.python.eager import execute as _execute
from tensorflow.python.framework import dtypes as _dtypes
from tensorflow.python.framework import errors as _errors
from tensorflow.python.framework import tensor_shape as _tensor_shape

from tensorflow.core.framework import op_def_pb2 as _op_def_pb2
# Needed to trigger the call to _set_call_cpp_shape_fn.
from tensorflow.python.framework import common_shapes as _common_shapes
from tensorflow.python.framework import op_def_registry as _op_def_registry
from tensorflow.python.framework import ops as _ops
from tensorflow.python.framework import op_def_library as _op_def_library
from tensorflow.python.util.deprecation import deprecated_endpoints
from tensorflow.python.util import dispatch as _dispatch
from tensorflow.python.util.tf_export import tf_export
from tensorflow.python.util.tf_export import kwarg_only as _kwarg_only
from tensorflow.tools.docs import doc_controls as _doc_controls


def encode_proto(sizes, values, field_names, message_type, descriptor_source="local://", name=None):
  r"""The op serializes protobuf messages provided in the input tensors.

  The types of the tensors in `values` must match the schema for the fields
  specified in `field_names`. All the tensors in `values` must have a common
  shape prefix, *batch_shape*.

  The `sizes` tensor specifies repeat counts for each field.  The repeat count
  (last dimension) of a each tensor in `values` must be greater than or equal
  to corresponding repeat count in `sizes`.

  A `message_type` name must be provided to give context for the field names.
  The actual message descriptor can be looked up either in the linked-in
  descriptor pool or a filename provided by the caller using the
  `descriptor_source` attribute.

  For the most part, the mapping between Proto field types and TensorFlow dtypes
  is straightforward. However, there are a few special cases:

  - A proto field that contains a submessage or group can only be converted
  to `DT_STRING` (the serialized submessage). This is to reduce the complexity
  of the API. The resulting string can be used as input to another instance of
  the decode_proto op.

  - TensorFlow lacks support for unsigned integers. The ops represent uint64
  types as a `DT_INT64` with the same twos-complement bit pattern (the obvious
  way). Unsigned int32 values can be represented exactly by specifying type
  `DT_INT64`, or using twos-complement if the caller specifies `DT_INT32` in
  the `output_types` attribute.

  The `descriptor_source` attribute selects the source of protocol
  descriptors to consult when looking up `message_type`. This may be:

  - An empty string  or "local://", in which case protocol descriptors are
  created for C++ (not Python) proto definitions linked to the binary.

  - A file, in which case protocol descriptors are created from the file,
  which is expected to contain a `FileDescriptorSet` serialized as a string.
  NOTE: You can build a `descriptor_source` file using the `--descriptor_set_out`
  and `--include_imports` options to the protocol compiler `protoc`.

  - A "bytes://<bytes>", in which protocol descriptors are created from `<bytes>`,
  which is expected to be a `FileDescriptorSet` serialized as a string.

  Args:
    sizes: A `Tensor` of type `int32`.
      Tensor of int32 with shape `[batch_shape, len(field_names)]`.
    values: A list of `Tensor` objects.
      List of tensors containing values for the corresponding field.
    field_names: A list of `strings`.
      List of strings containing proto field names.
    message_type: A `string`. Name of the proto message type to decode.
    descriptor_source: An optional `string`. Defaults to `"local://"`.
    name: A name for the operation (optional).

  Returns:
    A `Tensor` of type `string`.
  """
  _ctx = _context._context or _context.context()
  if _ctx is not None and _ctx._thread_local_data.is_eager:
    try:
      _result = _pywrap_tensorflow.TFE_Py_FastPathExecute(
        _ctx._context_handle, _ctx._thread_local_data.device_name,
        "EncodeProto", name, _ctx.post_execution_callbacks, sizes, values,
        "field_names", field_names, "message_type", message_type,
        "descriptor_source", descriptor_source)
      return _result
    except _core._FallbackException:
      try:
        return encode_proto_eager_fallback(
            sizes, values, field_names=field_names, message_type=message_type,
            descriptor_source=descriptor_source, name=name, ctx=_ctx)
      except _core._SymbolicException:
        pass  # Add nodes to the TensorFlow graph.
    except _core._NotOkStatusException as e:
      if name is not None:
        message = e.message + " name: " + name
      else:
        message = e.message
      _six.raise_from(_core._status_to_exception(e.code, message), None)
  # Add nodes to the TensorFlow graph.
  if not isinstance(field_names, (list, tuple)):
    raise TypeError(
        "Expected list for 'field_names' argument to "
        "'encode_proto' Op, not %r." % field_names)
  field_names = [_execute.make_str(_s, "field_names") for _s in field_names]
  message_type = _execute.make_str(message_type, "message_type")
  if descriptor_source is None:
    descriptor_source = "local://"
  descriptor_source = _execute.make_str(descriptor_source, "descriptor_source")
  _, _, _op = _op_def_lib._apply_op_helper(
        "EncodeProto", sizes=sizes, values=values, field_names=field_names,
                       message_type=message_type,
                       descriptor_source=descriptor_source, name=name)
  _result = _op.outputs[:]
  _inputs_flat = _op.inputs
  _attrs = ("field_names", _op.get_attr("field_names"), "message_type",
            _op.get_attr("message_type"), "descriptor_source",
            _op.get_attr("descriptor_source"), "Tinput_types",
            _op.get_attr("Tinput_types"))
  _execute.record_gradient(
      "EncodeProto", _inputs_flat, _attrs, _result, name)
  _result, = _result
  return _result

def EncodeProto(sizes, values, field_names, message_type, descriptor_source="local://", name=None):
  return encode_proto(sizes=sizes, values=values, field_names=field_names, message_type=message_type, descriptor_source=descriptor_source, name=name)
EncodeProto.__doc__ = encode_proto.__doc__
EncodeProto = _doc_controls.do_not_generate_docs(_kwarg_only(EncodeProto))
tf_export("raw_ops.EncodeProto")(EncodeProto)


def encode_proto_eager_fallback(sizes, values, field_names, message_type, descriptor_source="local://", name=None, ctx=None):
  r"""This is the slowpath function for Eager mode.
  This is for function encode_proto
  """
  _ctx = ctx if ctx else _context.context()
  if not isinstance(field_names, (list, tuple)):
    raise TypeError(
        "Expected list for 'field_names' argument to "
        "'encode_proto' Op, not %r." % field_names)
  field_names = [_execute.make_str(_s, "field_names") for _s in field_names]
  message_type = _execute.make_str(message_type, "message_type")
  if descriptor_source is None:
    descriptor_source = "local://"
  descriptor_source = _execute.make_str(descriptor_source, "descriptor_source")
  _attr_Tinput_types, values = _execute.convert_to_mixed_eager_tensors(values, _ctx)
  sizes = _ops.convert_to_tensor(sizes, _dtypes.int32)
  _inputs_flat = [sizes] + list(values)
  _attrs = ("field_names", field_names, "message_type", message_type,
  "descriptor_source", descriptor_source, "Tinput_types", _attr_Tinput_types)
  _result = _execute.execute(b"EncodeProto", 1, inputs=_inputs_flat,
                             attrs=_attrs, ctx=_ctx, name=name)
  _execute.record_gradient(
      "EncodeProto", _inputs_flat, _attrs, _result, name)
  _result, = _result
  return _result

def _InitOpDefLibrary(op_list_proto_bytes):
  op_list = _op_def_pb2.OpList()
  op_list.ParseFromString(op_list_proto_bytes)
  _op_def_registry.register_op_list(op_list)
  op_def_lib = _op_def_library.OpDefLibrary()
  op_def_lib.add_op_list(op_list)
  return op_def_lib
# op {
#   name: "EncodeProto"
#   input_arg {
#     name: "sizes"
#     type: DT_INT32
#   }
#   input_arg {
#     name: "values"
#     type_list_attr: "Tinput_types"
#   }
#   output_arg {
#     name: "bytes"
#     type: DT_STRING
#   }
#   attr {
#     name: "field_names"
#     type: "list(string)"
#   }
#   attr {
#     name: "message_type"
#     type: "string"
#   }
#   attr {
#     name: "descriptor_source"
#     type: "string"
#     default_value {
#       s: "local://"
#     }
#   }
#   attr {
#     name: "Tinput_types"
#     type: "list(type)"
#     has_minimum: true
#     minimum: 1
#   }
# }
_op_def_lib = _InitOpDefLibrary(b"\n\271\001\n\013EncodeProto\022\t\n\005sizes\030\003\022\026\n\006values2\014Tinput_types\032\t\n\005bytes\030\007\"\033\n\013field_names\022\014list(string)\"\026\n\014message_type\022\006string\"\'\n\021descriptor_source\022\006string\032\n\022\010local://\"\036\n\014Tinput_types\022\nlist(type)(\0010\001")



# Copyright 2016 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for loading op libraries.

@@load_op_library
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import re

from tensorflow.python.framework import load_library
from tensorflow.python.platform import resource_loader


def load_op_library(path):
  """Loads a contrib op library from the given path.

  NOTE(mrry): On Windows, we currently assume that some contrib op
  libraries are statically linked into the main TensorFlow Python
  extension DLL - use dynamically linked ops if the .so is present.

  Args:
    path: An absolute path to a shared object file.

  Returns:
    A Python module containing the Python wrappers for Ops defined in the
    plugin.
  """
  if os.name == 'nt':
    # To avoid making every user_ops aware of windows, re-write
    # the file extension from .so to .dll if .so file doesn't exist.
    if not os.path.exists(path):
      path = re.sub(r'\.so$', '.dll', path)

    # Currently we have only some user_ops as dlls on windows - don't try
    # to load them if the dll is not found.
    # TODO(mrry): Once we have all of them this check should be removed.
    if not os.path.exists(path):
      return None
  path = resource_loader.get_path_to_datafile(path)
  ret = load_library.load_op_library(path)
  assert ret, 'Could not load %s' % path
  return ret



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Built-in loss functions.

"""

from __future__ import print_function as _print_function

import sys as _sys

from tensorflow.python.keras.losses import BinaryCrossentropy
from tensorflow.python.keras.losses import CategoricalCrossentropy
from tensorflow.python.keras.losses import CategoricalHinge
from tensorflow.python.keras.losses import CosineSimilarity
from tensorflow.python.keras.losses import Hinge
from tensorflow.python.keras.losses import Huber
from tensorflow.python.keras.losses import KLD
from tensorflow.python.keras.losses import KLD as kld
from tensorflow.python.keras.losses import KLD as kullback_leibler_divergence
from tensorflow.python.keras.losses import KLDivergence
from tensorflow.python.keras.losses import LogCosh
from tensorflow.python.keras.losses import Loss
from tensorflow.python.keras.losses import MAE
from tensorflow.python.keras.losses import MAE as mae
from tensorflow.python.keras.losses import MAE as mean_absolute_error
from tensorflow.python.keras.losses import MAPE
from tensorflow.python.keras.losses import MAPE as mape
from tensorflow.python.keras.losses import MAPE as mean_absolute_percentage_error
from tensorflow.python.keras.losses import MSE
from tensorflow.python.keras.losses import MSE as mean_squared_error
from tensorflow.python.keras.losses import MSE as mse
from tensorflow.python.keras.losses import MSLE
from tensorflow.python.keras.losses import MSLE as mean_squared_logarithmic_error
from tensorflow.python.keras.losses import MSLE as msle
from tensorflow.python.keras.losses import MeanAbsoluteError
from tensorflow.python.keras.losses import MeanAbsolutePercentageError
from tensorflow.python.keras.losses import MeanSquaredError
from tensorflow.python.keras.losses import MeanSquaredLogarithmicError
from tensorflow.python.keras.losses import Poisson
from tensorflow.python.keras.losses import SparseCategoricalCrossentropy
from tensorflow.python.keras.losses import SquaredHinge
from tensorflow.python.keras.losses import binary_crossentropy
from tensorflow.python.keras.losses import categorical_crossentropy
from tensorflow.python.keras.losses import categorical_hinge
from tensorflow.python.keras.losses import cosine_proximity
from tensorflow.python.keras.losses import cosine_proximity as cosine
from tensorflow.python.keras.losses import cosine_proximity as cosine_similarity
from tensorflow.python.keras.losses import deserialize
from tensorflow.python.keras.losses import get
from tensorflow.python.keras.losses import hinge
from tensorflow.python.keras.losses import logcosh
from tensorflow.python.keras.losses import poisson
from tensorflow.python.keras.losses import serialize
from tensorflow.python.keras.losses import sparse_categorical_crossentropy
from tensorflow.python.keras.losses import squared_hinge

del _print_function

from tensorflow.python.util import module_wrapper as _module_wrapper

if not isinstance(_sys.modules[__name__], _module_wrapper.TFModuleWrapper):
  _sys.modules[__name__] = _module_wrapper.TFModuleWrapper(
      _sys.modules[__name__], "keras.losses", public_apis=None, deprecation=True,
      has_lite=False)



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Provides a list of renames between TensorFlow 1.* and 2.0."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.tools.compatibility import renames_v2

# pylint: disable=line-too-long

# Add additional renames not in renames_v2.py here.
# IMPORTANT: For the renames in here, if you also need to add to
# function_reorders or function_keyword_renames in tf_upgrade_v2.py,
# use the OLD function name.
# These renames happen after the arguments have been processed.
manual_symbol_renames = {
    "tf.batch_to_space_nd":
        "tf.batch_to_space",
    "tf.batch_gather":
        "tf.compat.v1.batch_gather",
    "tf.space_to_batch_nd":
        "tf.space_to_batch",
    "tf.nn.space_to_batch":
        "tf.space_to_batch",
    "tf.estimator.inputs":
        "tf.compat.v1.estimator.inputs",
    "tf.extract_image_patches":
        "tf.image.extract_patches",
    "tf.image.extract_image_patches":
        "tf.image.extract_patches",
    "tf.gfile.Copy":
        "tf.io.gfile.copy",
    "tf.gfile.DeleteRecursively":
        "tf.io.gfile.rmtree",
    "tf.gfile.Exists":
        "tf.io.gfile.exists",
    "tf.gfile.Glob":
        "tf.io.gfile.glob",
    "tf.gfile.GFile":
        "tf.io.gfile.GFile",
    "tf.gfile.IsDirectory":
        "tf.io.gfile.isdir",
    "tf.gfile.ListDirectory":
        "tf.io.gfile.listdir",
    "tf.gfile.MakeDirs":
        "tf.io.gfile.makedirs",
    "tf.gfile.MkDir":
        "tf.io.gfile.mkdir",
    "tf.gfile.Open":
        "tf.io.gfile.GFile",
    "tf.gfile.Remove":
        "tf.io.gfile.remove",
    "tf.gfile.Rename":
        "tf.io.gfile.rename",
    "tf.gfile.Stat":
        "tf.io.gfile.stat",
    "tf.gfile.Walk":
        "tf.io.gfile.walk",
    "tf.contrib.cluster_resolver.ClusterResolver":
        "tf.distribute.cluster_resolver.ClusterResolver",
    "tf.contrib.cluster_resolver.GceClusterResolver":
        "tf.distribute.cluster_resolver.GCEClusterResolver",
    "tf.contrib.cluster_resolver.KubernetesClusterResolver":
        "tf.distribute.cluster_resolver.KubernetesClusterResolver",
    "tf.contrib.cluster_resolver.SimpleClusterResolver":
        "tf.distribute.cluster_resolver.SimpleClusterResolver",
    "tf.contrib.cluster_resolver.SlurmClusterResolver":
        "tf.distribute.cluster_resolver.SlurmClusterResolver",
    "tf.contrib.cluster_resolver.TFConfigClusterResolver":
        "tf.distribute.cluster_resolver.TFConfigClusterResolver",
    "tf.contrib.cluster_resolver.TPUClusterResolver":
        "tf.distribute.cluster_resolver.TPUClusterResolver",
    "tf.contrib.cluster_resolver.UnionClusterResolver":
        "tf.distribute.cluster_resolver.UnionClusterResolver",
    "tf.contrib.data.AUTOTUNE":
        "tf.data.experimental.AUTOTUNE",
    "tf.contrib.data.Counter":
        "tf.data.experimental.Counter",
    "tf.contrib.data.CheckpointInputPipelineHook":
        "tf.data.experimental.CheckpointInputPipelineHook",
    "tf.contrib.data.CsvDataset":
        "tf.data.experimental.CsvDataset",
    "tf.contrib.data.Optional":
        "tf.data.experimental.Optional",
    "tf.contrib.data.RandomDataset":
        "tf.data.experimental.RandomDataset",
    "tf.contrib.data.Reducer":
        "tf.data.experimental.Reducer",
    "tf.contrib.data.SqlDataset":
        "tf.data.experimental.SqlDataset",
    "tf.contrib.data.StatsAggregator":
        "tf.data.experimental.StatsAggregator",
    "tf.contrib.data.TFRecordWriter":
        "tf.data.experimental.TFRecordWriter",
    "tf.contrib.data.assert_element_shape":
        "tf.data.experimental.assert_element_shape",
    "tf.contrib.data.bucket_by_sequence_length":
        "tf.data.experimental.bucket_by_sequence_length",
    "tf.contrib.data.choose_from_datasets":
        "tf.data.experimental.choose_from_datasets",
    "tf.contrib.data.copy_to_device":
        "tf.data.experimental.copy_to_device",
    "tf.contrib.data.dense_to_sparse_batch":
        "tf.data.experimental.dense_to_sparse_batch",
    "tf.contrib.data.enumerate_dataset":
        "tf.data.experimental.enumerate_dataset",
    "tf.contrib.data.get_next_as_optional":
        "tf.data.experimental.get_next_as_optional",
    "tf.contrib.data.get_single_element":
        "tf.data.experimental.get_single_element",
    "tf.contrib.data.group_by_reducer":
        "tf.data.experimental.group_by_reducer",
    "tf.contrib.data.group_by_window":
        "tf.data.experimental.group_by_window",
    "tf.contrib.data.ignore_errors":
        "tf.data.experimental.ignore_errors",
    "tf.contrib.data.latency_stats":
        "tf.data.experimental.latency_stats",
    "tf.contrib.data.make_batched_features_dataset":
        "tf.data.experimental.make_batched_features_dataset",
    "tf.contrib.data.make_csv_dataset":
        "tf.data.experimental.make_csv_dataset",
    "tf.contrib.data.make_saveable_from_iterator":
        "tf.data.experimental.make_saveable_from_iterator",
    "tf.contrib.data.map_and_batch":
        "tf.data.experimental.map_and_batch",
    "tf.contrib.data.parallel_interleave":
        "tf.data.experimental.parallel_interleave",
    "tf.contrib.data.parse_example_dataset":
        "tf.data.experimental.parse_example_dataset",
    "tf.contrib.data.prefetch_to_device":
        "tf.data.experimental.prefetch_to_device",
    "tf.contrib.data.rejection_resample":
        "tf.data.experimental.rejection_resample",
    "tf.contrib.data.sample_from_datasets":
        "tf.data.experimental.sample_from_datasets",
    "tf.contrib.data.scan":
        "tf.data.experimental.scan",
    "tf.contrib.data.set_stats_aggregator":
        "tf.data.experimental.set_stats_aggregator",
    "tf.contrib.data.shuffle_and_repeat":
        "tf.data.experimental.shuffle_and_repeat",
    "tf.contrib.data.unbatch":
        "tf.data.experimental.unbatch",
    "tf.contrib.data.unique":
        "tf.data.experimental.unique",
    "tf.contrib.distribute.CrossDeviceOps":
        "tf.distribute.CrossDeviceOps",
    "tf.contrib.distribute.ReductionToOneDeviceCrossDeviceOps":
        "tf.distribute.ReductionToOneDevice",
    "tf.contrib.estimator.make_early_stopping_hook":
        "tf.estimator.experimental.make_early_stopping_hook",
    "tf.contrib.estimator.stop_if_higher_hook":
        "tf.estimator.experimental.stop_if_higher_hook",
    "tf.contrib.estimator.stop_if_lower_hook":
        "tf.estimator.experimental.stop_if_lower_hook",
    "tf.contrib.estimator.stop_if_no_decrease_hook":
        "tf.estimator.experimental.stop_if_no_decrease_hook",
    "tf.contrib.estimator.stop_if_no_increase_hook":
        "tf.estimator.experimental.stop_if_no_increase_hook",
    "tf.contrib.framework.CriticalSection":
        "tf.CriticalSection",
    "tf.contrib.framework.is_tensor":
        "tf.is_tensor",
    "tf.contrib.framework.load_variable":
        "tf.train.load_variable",
    "tf.contrib.framework.nest.assert_same_structure":
        "tf.nest.assert_same_structure",
    "tf.contrib.framework.nest.flatten":
        "tf.nest.flatten",
    "tf.contrib.framework.nest.is_sequence":
        "tf.nest.is_nested",
    "tf.contrib.framework.nest.map_structure":
        "tf.nest.map_structure",
    "tf.contrib.framework.nest.pack_sequence_as":
        "tf.nest.pack_sequence_as",
    "tf.contrib.batching.batch_function":
        "tf.nondifferentiable_batch_function",
    "tf.contrib.util.constant_value":
        "tf.get_static_value",
    "tf.contrib.saved_model.load_keras_model":
        "tf.keras.experimental.load_from_saved_model",
    "tf.contrib.saved_model.save_keras_model":
        "tf.keras.experimental.export_saved_model",
    "tf.contrib.rnn.RNNCell":
        "tf.compat.v1.nn.rnn_cell.RNNCell",
    "tf.contrib.rnn.LSTMStateTuple":
        "tf.nn.rnn_cell.LSTMStateTuple",
    "tf.contrib.rnn.BasicLSTMCell":
        "tf.compat.v1.nn.rnn_cell.BasicLSTMCell",
    "tf.contrib.rnn.BasicRNNCell":
        "tf.compat.v1.nn.rnn_cell.BasicRNNCell",
    "tf.contrib.rnn.GRUCell":
        "tf.compat.v1.nn.rnn_cell.GRUCell",
    "tf.contrib.rnn.LSTMCell":
        "tf.compat.v1.nn.rnn_cell.LSTMCell",
    "tf.contrib.rnn.MultiRNNCell":
        "tf.compat.v1.nn.rnn_cell.MultiRNNCell",
    "tf.contrib.rnn.static_rnn":
        "tf.compat.v1.nn.static_rnn",
    "tf.contrib.rnn.static_state_saving_rnn":
        "tf.compat.v1.nn.static_state_saving_rnn",
    "tf.contrib.rnn.static_bidirectional_rnn":
        "tf.compat.v1.nn.static_bidirectional_rnn",
    "tf.contrib.framework.sort":
        "tf.sort",
    "tf.contrib.framework.argsort":
        "tf.argsort",
    "tf.contrib.summary.all_summary_ops":
        "tf.compat.v1.summary.all_v2_summary_ops",
    "tf.contrib.summary.always_record_summaries":
        "tf.compat.v2.summary.record_if",
    "tf.contrib.summary.audio":
        "tf.compat.v2.summary.audio",
    "tf.contrib.summary.create_file_writer":
        "tf.compat.v2.summary.create_file_writer",
    "tf.contrib.summary.flush":
        "tf.compat.v2.summary.flush",
    "tf.contrib.summary.generic":
        "tf.compat.v2.summary.write",
    "tf.contrib.summary.histogram":
        "tf.compat.v2.summary.histogram",
    "tf.contrib.summary.image":
        "tf.compat.v2.summary.image",
    "tf.contrib.summary.initialize":
        "tf.compat.v1.summary.initialize",
    "tf.contrib.summary.never_record_summaries":
        "tf.compat.v2.summary.record_if",
    "tf.contrib.summary.scalar":
        "tf.compat.v2.summary.scalar",
    "tf.contrib.tpu.CrossShardOptimizer":
        "tf.compat.v1.tpu.CrossShardOptimizer",
    "tf.contrib.tpu.InputPipelineConfig":
        "tf.compat.v1.estimator.tpu.InputPipelineConfig",
    "tf.contrib.tpu.RunConfig":
        "tf.compat.v1.estimator.tpu.RunConfig",
    "tf.contrib.tpu.TPUConfig":
        "tf.compat.v1.estimator.tpu.TPUConfig",
    "tf.contrib.tpu.TPUEstimator":
        "tf.compat.v1.estimator.tpu.TPUEstimator",
    "tf.contrib.tpu.TPUEstimatorSpec":
        "tf.compat.v1.estimator.tpu.TPUEstimatorSpec",
    "tf.contrib.tpu.batch_parallel":
        "tf.compat.v1.tpu.batch_parallel",
    "tf.contrib.tpu.bfloat16_scope":
        "tf.compat.v1.tpu.bfloat16_scope",
    "tf.contrib.tpu.core":
        "tf.compat.v1.tpu.core",
    "tf.contrib.tpu.cross_replica_sum":
        "tf.compat.v1.tpu.cross_replica_sum",
    "tf.contrib.tpu.initialize_system":
        "tf.compat.v1.tpu.initialize_system",
    "tf.contrib.tpu.outside_compilation":
        "tf.compat.v1.tpu.outside_compilation",
    "tf.contrib.tpu.replicate":
        "tf.compat.v1.tpu.replicate",
    "tf.contrib.tpu.rewrite":
        "tf.compat.v1.tpu.rewrite",
    "tf.contrib.tpu.shard":
        "tf.compat.v1.tpu.shard",
    "tf.contrib.tpu.shutdown_system":
        "tf.compat.v1.tpu.shutdown_system",
    "tf.contrib.training.checkpoints_iterator":
        "tf.train.checkpoints_iterator",
    "tf.contrib.layers.recompute_grad":
        "tf.recompute_grad",
    "tf.count_nonzero":
        "tf.math.count_nonzero",
    "tf.manip.batch_to_space_nd":
        "tf.batch_to_space",
    "tf.quantize_v2":
        "tf.quantization.quantize",
    "tf.sparse_add":
        "tf.sparse.add",
    "tf.sparse_concat":
        "tf.sparse.concat",
    "tf.sparse_split":
        "tf.sparse.split",
    "tf.sparse_matmul":
        "tf.linalg.matmul",
    "tf.sparse_reduce_sum":
        "tf.sparse.reduce_sum",
    "tf.sparse_reduce_max":
        "tf.sparse.reduce_max",
    "tf.random.stateless_multinomial":
        "tf.random.stateless_categorical",
    "tf.substr":
        "tf.strings.substr",
    # TODO(b/129398290)
    "tf.string_split":
        "tf.compat.v1.string_split",
    "tf.string_to_hash_bucket":
        "tf.strings.to_hash_bucket",
    "tf.string_to_number":
        "tf.strings.to_number",
    "tf.multinomial":
        "tf.random.categorical",
    "tf.random.multinomial":
        "tf.random.categorical",
    "tf.reduce_join":
        "tf.strings.reduce_join",
    "tf.load_file_system_library":
        "tf.load_library",
    "tf.bincount":
        "tf.math.bincount",
    "tf.confusion_matrix":
        "tf.math.confusion_matrix",
    "tf.train.confusion_matrix":
        "tf.math.confusion_matrix",
    "tf.train.sdca_fprint":
        "tf.raw_ops.SdcaFprint",
    "tf.train.sdca_optimizer":
        "tf.raw_ops.SdcaOptimizer",
    "tf.train.sdca_shrink_l1":
        "tf.raw_ops.SdcaShrinkL1",
    "tf.decode_csv":
        "tf.io.decode_csv",
    "tf.data.Iterator":
        "tf.compat.v1.data.Iterator",
    "tf.data.experimental.DatasetStructure":
        "tf.data.DatasetSpec",
    "tf.data.experimental.OptionalStructure":
        "tf.OptionalSpec",
    "tf.data.experimental.RaggedTensorStructure":
        "tf.RaggedTensorSpec",
    "tf.data.experimental.SparseTensorStructure":
        "tf.SparseTensorSpec",
    "tf.data.experimental.Structure":
        "tf.TypeSpec",
    "tf.data.experimental.TensorArrayStructure":
        "tf.TensorArraySpec",
    "tf.data.experimental.TensorStructure":
        "tf.TensorSpec",
    "tf.parse_example":
        "tf.io.parse_example",
    "tf.parse_single_example":
        "tf.io.parse_single_example",
    "tf.nn.fused_batch_norm":
        "tf.compat.v1.nn.fused_batch_norm",
    "tf.nn.softmax_cross_entropy_with_logits_v2":
        "tf.nn.softmax_cross_entropy_with_logits",
    "tf.losses.Reduction.MEAN":
        "tf.compat.v1.losses.Reduction.MEAN",
    "tf.losses.Reduction.SUM_BY_NONZERO_WEIGHTS":
        "tf.compat.v1.losses.Reduction.SUM_BY_NONZERO_WEIGHTS",
    "tf.losses.Reduction.SUM_OVER_NONZERO_WEIGHTS":
        "tf.compat.v1.losses.Reduction.SUM_OVER_NONZERO_WEIGHTS",
    "tf.lite.constants.FLOAT":
        "tf.float32",
    "tf.lite.constants.INT32":
        "tf.int32",
    "tf.lite.constants.INT64":
        "tf.int64",
    "tf.lite.constants.STRING":
        "tf.string",
    "tf.lite.constants.QUANTIZED_UINT8":
        "tf.uint8",
    "tf.arg_max":
        "tf.argmax",
    "tf.arg_min":
        "tf.argmin",
    # tf.nn.ctc_loss is still available in 2.0 but behavior
    # changed significantly.
    "tf.nn.ctc_loss":
        "tf.compat.v1.nn.ctc_loss",
    # tf.saved_model.load in 1.x has no equivalent in 2.x, but there is a
    # symbol with the same name.
    "tf.saved_model.load":
        "tf.compat.v1.saved_model.load",
    "tf.saved_model.load_v2":
        "tf.compat.v2.saved_model.load",
    "tf.image.resize_images":
        "tf.image.resize",
    "tf.random_poisson":
        "tf.random.poisson",
    "tf.debugging.assert_greater":
        "tf.compat.v1.debugging.assert_greater",
    "tf.debugging.assert_greater_equal":
        "tf.compat.v1.debugging.assert_greater_equal",
    "tf.debugging.assert_integer":
        "tf.compat.v1.debugging.assert_integer",
    "tf.debugging.assert_less":
        "tf.compat.v1.debugging.assert_less",
    "tf.debugging.assert_less_equal":
        "tf.compat.v1.debugging.assert_less_equal",
    "tf.debugging.assert_near":
        "tf.compat.v1.debugging.assert_near",
    "tf.debugging.assert_negative":
        "tf.compat.v1.debugging.assert_negative",
    "tf.debugging.assert_non_negative":
        "tf.compat.v1.debugging.assert_non_negative",
    "tf.debugging.assert_non_positive":
        "tf.compat.v1.debugging.assert_non_positive",
    "tf.debugging.assert_none_equal":
        "tf.compat.v1.debugging.assert_none_equal",
    "tf.debugging.assert_type":
        "tf.compat.v1.debugging.assert_type",
    "tf.debugging.assert_positive":
        "tf.compat.v1.debugging.assert_positive",
    "tf.debugging.assert_equal":
        "tf.compat.v1.debugging.assert_equal",
    "tf.debugging.assert_scalar":
        "tf.compat.v1.debugging.assert_scalar",
    "tf.assert_equal":
        "tf.compat.v1.assert_equal",
    "tf.assert_less":
        "tf.compat.v1.assert_less",
    "tf.assert_greater":
        "tf.compat.v1.assert_greater",
    "tf.debugging.assert_rank":
        "tf.compat.v1.debugging.assert_rank",
    "tf.debugging.assert_rank_at_least":
        "tf.compat.v1.debugging.assert_rank_at_least",
    "tf.debugging.assert_rank_in":
        "tf.compat.v1.debugging.assert_rank_in",
    "tf.errors.exception_type_from_error_code":
        "tf.compat.v1.errors.exception_type_from_error_code",
    "tf.errors.error_code_from_exception_type":
        "tf.compat.v1.errors.error_code_from_exception_type",
    "tf.errors.raise_exception_on_not_ok_status":
        "tf.compat.v1.errors.raise_exception_on_not_ok_status",
    "tf.assert_rank":
        "tf.compat.v1.assert_rank",
    "tf.nn.max_pool":
        "tf.nn.max_pool2d",
    "tf.nn.avg_pool":
        "tf.nn.avg_pool2d",
    "tf.keras.initializers.zeros":
        "tf.compat.v1.keras.initializers.zeros",
    "tf.keras.initializers.Zeros":
        "tf.compat.v1.keras.initializers.Zeros",
    "tf.keras.initializers.ones":
        "tf.compat.v1.keras.initializers.ones",
    "tf.keras.initializers.Ones":
        "tf.compat.v1.keras.initializers.Ones",
    "tf.keras.initializers.constant":
        "tf.compat.v1.keras.initializers.constant",
    "tf.keras.initializers.Constant":
        "tf.compat.v1.keras.initializers.Constant",
    "tf.keras.initializers.VarianceScaling":
        "tf.compat.v1.keras.initializers.VarianceScaling",
    "tf.keras.initializers.Orthogonal":
        "tf.compat.v1.keras.initializers.Orthogonal",
    "tf.keras.initializers.orthogonal":
        "tf.compat.v1.keras.initializers.orthogonal",
    "tf.keras.initializers.Identity":
        "tf.compat.v1.keras.initializers.Identity",
    "tf.keras.initializers.identity":
        "tf.compat.v1.keras.initializers.identity",
    "tf.keras.initializers.glorot_uniform":
        "tf.compat.v1.keras.initializers.glorot_uniform",
    "tf.keras.initializers.glorot_normal":
        "tf.compat.v1.keras.initializers.glorot_normal",
    "tf.keras.initializers.lecun_normal":
        "tf.compat.v1.keras.initializers.lecun_normal",
    "tf.keras.initializers.lecun_uniform":
        "tf.compat.v1.keras.initializers.lecun_uniform",
    "tf.keras.initializers.he_normal":
        "tf.compat.v1.keras.initializers.he_normal",
    "tf.keras.initializers.he_uniform":
        "tf.compat.v1.keras.initializers.he_uniform",
    "tf.keras.initializers.TruncatedNormal":
        "tf.compat.v1.keras.initializers.TruncatedNormal",
    "tf.keras.initializers.truncated_normal":
        "tf.compat.v1.keras.initializers.truncated_normal",
    "tf.keras.initializers.RandomUniform":
        "tf.compat.v1.keras.initializers.RandomUniform",
    "tf.keras.initializers.uniform":
        "tf.compat.v1.keras.initializers.uniform",
    "tf.keras.initializers.random_uniform":
        "tf.compat.v1.keras.initializers.random_uniform",
    "tf.keras.initializers.RandomNormal":
        "tf.compat.v1.keras.initializers.RandomNormal",
    "tf.keras.initializers.normal":
        "tf.compat.v1.keras.initializers.normal",
    "tf.keras.initializers.random_normal":
        "tf.compat.v1.keras.initializers.random_normal",
    "tf.zeros_initializer":
        "tf.compat.v1.zeros_initializer",
    "tf.initializers.zeros":
        "tf.compat.v1.initializers.zeros",
    "tf.ones_initializer":
        "tf.compat.v1.ones_initializer",
    "tf.initializers.ones":
        "tf.compat.v1.initializers.ones",
    "tf.constant_initializer":
        "tf.compat.v1.constant_initializer",
    "tf.initializers.constant":
        "tf.compat.v1.initializers.constant",
    "tf.random_uniform_initializer":
        "tf.compat.v1.random_uniform_initializer",
    "tf.initializers.random_uniform":
        "tf.compat.v1.initializers.random_uniform",
    "tf.random_normal_initializer":
        "tf.compat.v1.random_normal_initializer",
    "tf.initializers.random_normal":
        "tf.compat.v1.initializers.random_normal",
    "tf.truncated_normal_initializer":
        "tf.compat.v1.truncated_normal_initializer",
    "tf.initializers.truncated_normal":
        "tf.compat.v1.initializers.truncated_normal",
    "tf.variance_scaling_initializer":
        "tf.compat.v1.variance_scaling_initializer",
    "tf.initializers.variance_scaling":
        "tf.compat.v1.initializers.variance_scaling",
    "tf.orthogonal_initializer":
        "tf.compat.v1.orthogonal_initializer",
    "tf.initializers.orthogonal":
        "tf.compat.v1.initializers.orthogonal",
    "tf.glorot_uniform_initializer":
        "tf.compat.v1.glorot_uniform_initializer",
    "tf.initializers.glorot_uniform":
        "tf.compat.v1.initializers.glorot_uniform",
    "tf.glorot_normal_initializer":
        "tf.compat.v1.glorot_normal_initializer",
    "tf.initializers.glorot_normal":
        "tf.compat.v1.initializers.glorot_normal",
    "tf.initializers.identity":
        "tf.compat.v1.initializers.identity",
    "tf.initializers.lecun_normal":
        "tf.compat.v1.initializers.lecun_normal",
    "tf.initializers.lecun_uniform":
        "tf.compat.v1.initializers.lecun_uniform",
    "tf.initializers.he_normal":
        "tf.compat.v1.initializers.he_normal",
    "tf.initializers.he_uniform":
        "tf.compat.v1.initializers.he_uniform",
    "tf.data.experimental.map_and_batch_with_legacy_function":
        "tf.compat.v1.data.experimental.map_and_batch_with_legacy_function",
    "tf.nn.conv2d_backprop_input":
        "tf.nn.conv2d_transpose",
    "tf.test.compute_gradient":
        "tf.compat.v1.test.compute_gradient",
    "tf.floor_div":
        "tf.math.floordiv",
    "tf.where":
        "tf.compat.v1.where",
    "tf.where_v2":
        "tf.compat.v2.where",
}
# pylint: enable=line-too-long

symbol_renames = renames_v2.renames
symbol_renames.update(manual_symbol_renames)

addons_symbol_mappings = {
    "tf.contrib.layers.poincare_normalize":
        "tfa.layers.PoincareNormalize",
    "tf.contrib.layers.maxout":
        "tfa.layers.Maxout",
    "tf.contrib.layers.group_norm":
        "tfa.layers.GroupNormalization",
    "tf.contrib.layers.instance_norm":
        "tfa.layers.InstanceNormalization",
    "tf.contrib.sparsemax.sparsemax":
        "tfa.activations.sparsemax",
    "tf.contrib.losses.metric_learning.contrastive_loss":
        "tfa.losses.ContrastiveLoss",
    "tf.contrib.losses.metric_learning.lifted_struct_loss":
        "tfa.losses.LiftedStructLoss",
    "tf.contrib.sparsemax.sparsemax_loss":
        "tfa.losses.SparsemaxLoss",
    "tf.contrib.losses.metric_learning.triplet_semihard_loss":
        "tfa.losses.TripletSemiHardLoss",
    "tf.contrib.opt.LazyAdamOptimizer":
        "tfa.optimizers.LazyAdam",
    "tf.contrib.opt.MovingAverageOptimizer":
        "tfa.optimizers.MovingAverage",
    "tf.contrib.opt.MomentumWOptimizer":
        "tfa.optimizers.SGDW",
    "tf.contrib.opt.AdamWOptimizer":
        "tfa.optimizers.AdamW",
    "tf.contrib.opt.extend_with_decoupled_weight_decay":
        "tfa.optimizers.extend_with_decoupled_weight_decay",
    "tf.contrib.text.skip_gram_sample":
        "tfa.text.skip_gram_sample",
    "tf.contrib.text.skip_gram_sample_with_text_vocab":
        "tfa.text.skip_gram_sample_with_text_vocab",
    "tf.contrib.image.dense_image_warp":
        "tfa.image.dense_image_warp",
    "tf.contrib.image.adjust_hsv_in_yiq":
        "tfa.image.adjust_hsv_in_yiq",
    "tf.contrib.image.compose_transforms":
        "tfa.image.compose_transforms",
    "tf.contrib.image.random_hsv_in_yiq":
        "tfa.image.random_hsv_in_yiq",
    "tf.contrib.image.angles_to_projective_transforms":
        "tfa.image.angles_to_projective_transforms",
    "tf.contrib.image.matrices_to_flat_transforms":
        "tfa.image.matricies_to_flat_transforms",
    "tf.contrib.image.rotate":
        "tfa.image.rotate",
    "tf.contrib.image.transform":
        "tfa.image.transform",
    "tf.contrib.rnn.NASCell":
        "tfa.rnn.NASCell",
    "tf.contrib.rnn.LayerNormBasicLSTMCell":
        "tfa.rnn.LayerNormLSTMCell"
}



# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Control flow statements: loops, conditionals, etc.

Note: most of these operators accept pairs of get_state/set_state functions, to
capture mutations that the corresponding code blocks might make. These
mutations only need to be captured when staging the control flow, and they just
work when reverting to Python behavior.

__Examples__

```
while cond:
  self.x += i
```

When the functionalized version is executed as a Python loop, it just works:

```
def loop_body():
  self.x += i     # works as expected for Python loops
```

But it won't work for TF loops:

```
def loop_body():
  self.x += i     # self.x has the wrong value!
```

get_state/set_state allow piping the mutations through the loop variables as
well, in effect changing the loop body:

```
def loop_body(self_x):
  self.x = self_x  # self.x now has the proper value
  self.x += i      # the original block
  self_x = self.x  # write self.x back into the loop vars
  return self_x

self_x = tf.while_loop(...)
self.x = self_x    # the result is not properly captured
```
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools
import numpy as np

from tensorflow.python.autograph.operators import py_builtins
from tensorflow.python.autograph.operators import special_values
from tensorflow.python.autograph.utils import ag_logging
from tensorflow.python.autograph.utils import misc
from tensorflow.python.autograph.utils import tensors
from tensorflow.python.data.experimental.ops import scan_ops
from tensorflow.python.data.experimental.ops import take_while_ops
from tensorflow.python.data.ops import dataset_ops
from tensorflow.python.data.ops import iterator_ops
from tensorflow.python.framework import constant_op
from tensorflow.python.framework import dtypes
from tensorflow.python.framework import func_graph
from tensorflow.python.framework import ops
from tensorflow.python.framework import tensor_util
from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import control_flow_util
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import tensor_array_ops
from tensorflow.python.util import nest

LIMIT_PYTHON_ITERATIONS = True
PYTHON_MAX_ITERATIONS = 100000000  # Fails in about one minute for empty loops.
WARN_INEFFICIENT_UNROLL = True
INEFFICIENT_UNROLL_MIN_ITERATIONS = 3000
INEFFICIENT_UNROLL_MIN_OPS = 1

def _disallow_undefs_into_loop(*values):
  """Ensures that all values in the state are defined when entering a loop."""
  undefined = tuple(filter(special_values.is_undefined, values))
  if undefined:
    raise ValueError(
        'TensorFlow requires that the following symbols must be defined'
        ' before the loop: {}'.format(tuple(s.symbol_name for s in undefined)))

  for value in values:
    if special_values.is_undefined_return(value):
      # Assumption: the loop will only capture the variable which tracks the
      # return value if the loop contained a return statement.
      # TODO(mdan): This should be checked at the place where return occurs.
      raise ValueError(
          'return statements are not supported within a TensorFlow loop.')


def _shape_greater_than_or_equal(shape1, shape2):
  """Check whether the shape2 is equal or more specific than shape1."""

  # The following logic was mirrored from control_flow_ops.py's
  # _ShapeLessThanOrEqual function.
  if shape1.dims is None:
    return True
  if shape1.ndims != shape2.ndims:
    return False
  for dim1, dim2 in zip(shape1.dims, shape2.dims):
    if dim1.value is not None and dim1.value != dim2.value:
      return False
  return True


def _verify_tf_loop_vars(init_loop_vars,
                         first_iter_vars,
                         basic_symbol_names,
                         composite_symbol_names,
                         include_shapes=True):
  """Verifies loop variables for consistency."""

  # The whole point of _verify_tf_loop_vars is to give more useful error message
  # than tf-level exception by including variable names.  If it's not available,
  # there is no point at performing this verification here.  As of 2019-07-31,
  # operators:control_flow_test does not pass the names.
  if basic_symbol_names is None:
    return

  output_symbol_names = basic_symbol_names + composite_symbol_names

  assert len(init_loop_vars) == len(first_iter_vars) == len(output_symbol_names)

  for init_loop_var, first_iter_var, name in zip(init_loop_vars,
                                                 first_iter_vars,
                                                 output_symbol_names):

    try:
      nest.assert_same_structure(
          init_loop_var, first_iter_var, expand_composites=True)
    except (ValueError, TypeError) as e:
      raise TypeError('"{}" does not have the same nested structure after one'
                      ' iteration.\n\n{}'.format(name, e))

    def _check_same_type(name, init_loop_var, first_iter_var):
      """Ensures init_loop_var and first_iter_var are consistent."""
      if isinstance(init_loop_var, (bool, int, float, str)):
        init_loop_var = ops.convert_to_tensor_v2(init_loop_var)

      if isinstance(first_iter_var, (bool, int, float, str)):
        first_iter_var = ops.convert_to_tensor_v2(first_iter_var)

      if (not tensor_util.is_tensor(init_loop_var) or
          not tensor_util.is_tensor(first_iter_var)):
        return

      # TODO(mdan): Properly account for CompositeTensors.
      if (not hasattr(init_loop_var, 'dtype') or
          not hasattr(first_iter_var, 'dtype')):
        return
      if (not hasattr(init_loop_var, 'shape') or
          not hasattr(first_iter_var, 'shape')):
        return

      if init_loop_var.dtype != first_iter_var.dtype:
        raise TypeError(
            '"{}" has dtype {} before the loop, but dtype {} after one'
            ' iteration. TensorFlow control flow requires it stays the'
            ' same.'.format(
                name,
                init_loop_var.dtype.name,
                first_iter_var.dtype.name,
            ))

      if include_shapes:
        init_shape = init_loop_var.shape
        first_iter_shape = first_iter_var.shape
        # TODO(b/135183013): Update needed once we support shape_invariants.
        if not _shape_greater_than_or_equal(init_shape, first_iter_shape):
          raise ValueError(
              '"{}" has shape {} before the loop, but shape {} after one'
              ' iteration. TensorFlow control flow requires it stays the'
              ' same or be more specific.'.format(name, init_shape,
                                                  first_iter_shape))

    nest.map_structure(
        functools.partial(_check_same_type, name), init_loop_var,
        first_iter_var)


def _verify_tf_cond_vars(body_outputs, orelse_outputs, basic_symbol_names,
                         composite_symbol_names):
  """Verifies variables manipulated by a conditional for consistency."""

  # The whole point of _verify_tf_cond_vars is to give more useful error message
  # than tf-level exception by including variable names.  If it's not available,
  # there is no point at performing this verification here.  As of 2019-07-31,
  # conditional expression does not pass the names.
  if basic_symbol_names is None:
    return

  output_symbol_names = basic_symbol_names + composite_symbol_names

  basic_body_outputs, composite_body_outputs = body_outputs
  basic_orelse_outputs, composite_orelse_outputs = orelse_outputs
  assert isinstance(composite_body_outputs, tuple)
  assert isinstance(composite_orelse_outputs, tuple)

  # TODO(kkimlabs): Make this more consistent.
  # The basic outputs should always be a tuple.
  if not isinstance(basic_body_outputs, tuple):
    basic_body_outputs = (basic_body_outputs,)
  if not isinstance(basic_orelse_outputs, tuple):
    basic_orelse_outputs = (basic_orelse_outputs,)

  body_outputs = basic_body_outputs + composite_body_outputs
  orelse_outputs = basic_orelse_outputs + composite_orelse_outputs

  for body_output, orelse_output, name in zip(body_outputs, orelse_outputs,
                                              output_symbol_names):
    try:
      nest.assert_same_structure(
          body_output, orelse_output, expand_composites=True)
    except (ValueError, TypeError) as e:
      raise TypeError(
          '"{}" does not have the same nested structure in the TRUE and FALSE'
          ' branches.\n\n{}'.format(name, str(e)))

    def _check_same_type(name, body_output_var, orelse_output_var):
      """Verfies that body_output_var and orelse_output_var have same dtype."""
      if isinstance(body_output_var, (bool, int, float, str)):
        body_output_var = ops.convert_to_tensor_v2(body_output_var)

      if isinstance(orelse_output_var, (bool, int, float, str)):
        orelse_output_var = ops.convert_to_tensor_v2(orelse_output_var)

      if (not tensor_util.is_tensor(body_output_var) or
          not tensor_util.is_tensor(orelse_output_var)):
        return

      # TODO(mdan): Properly account for CompositeTensors.
      if (not hasattr(body_output_var, 'dtype') or
          not hasattr(orelse_output_var, 'dtype')):
        return

      if body_output_var.dtype != orelse_output_var.dtype:
        raise TypeError(
            '"{}" has dtype {} in the TRUE branch, but dtype={} in the FALSE'
            ' branch. TensorFlow control flow requires that they are the'
            ' same.'.format(name, body_output_var.dtype.name,
                            orelse_output_var.dtype.name))

    nest.map_structure(
        functools.partial(_check_same_type, name), body_output, orelse_output)


def for_stmt(iter_,
             extra_test,
             body,
             get_state,
             set_state,
             init_vars,
             basic_symbol_names=None,
             composite_symbol_names=None):
  """Functional form of a for statement.

  The loop operates on a state, which includes all symbols that are
  variant across loop iterations, excluding the iterate as well as the
  variables local to the loop.

  For example, given the loop below that calculates the geometric and
  arithmetic means or some numbers:

    geo_mean = 1
    arith_mean = 0
    for i in range(n):
      a = numbers[i]
      geo_mean *= a
      arith_mean += a

  The state is represented by the variables geo_mean and arith_mean. The
  argument for initial_state may contain the tuple (1, 0), the body will
  include the arguments geo_mean and arith_mean and will return a tuple
  representing the new values for geo_mean and respectively arith_mean.

  Args:
    iter_: The entity being iterated over.
    extra_test: Callable with the state as arguments, and boolean return type.
      An additional loop condition.
    body: Callable with the iterate and the state as arguments, and state as
      return type. The actual loop body.
    get_state: Additional callable which can capture additional state (such as
      the values of composite symbols). This is only useful when staging the
      loop.
    set_state: Additional callable which save values captured by get_state back
      into the Python environment. This is only useful when staging the loop.
    init_vars: Tuple containing the initial state.
    basic_symbol_names: Tuple containing basic loop var names.
    composite_symbol_names: Tuple containing composite loop var names.

  Returns:
    Tuple containing the final state.
  """
  if tensor_util.is_tensor(iter_):
    if tensors.is_range_tensor(iter_):
      return _tf_range_for_stmt(iter_, extra_test, body, get_state, set_state,
                                init_vars, basic_symbol_names,
                                composite_symbol_names)
    else:
      return _known_len_tf_for_stmt(iter_, extra_test, body, get_state,
                                    set_state, init_vars, basic_symbol_names,
                                    composite_symbol_names)

  if isinstance(iter_, dataset_ops.DatasetV2):
    return _tf_dataset_for_stmt(iter_, extra_test, body, get_state, set_state,
                                init_vars, basic_symbol_names,
                                composite_symbol_names)

  if isinstance(iter_, iterator_ops.IteratorV2):
    return _tf_iterator_for_stmt(iter_, extra_test, body, get_state, set_state,
                                 init_vars, basic_symbol_names,
                                 composite_symbol_names)

  # Note: This experimental interface is subject to change.
  custom_handler = getattr(iter_, '_autograph_for_loop', None)
  if custom_handler is not None:
    # TODO(mdan): TensorFlow-specific verification - handlers should perform it.
    _disallow_undefs_into_loop(*init_vars)
    # TODO(mdan): Enable get_state/set_state separately.
    return custom_handler(extra_test, body, init_vars)

  return _py_for_stmt(iter_, extra_test, body, get_state, set_state, init_vars)


def _py_for_stmt(iter_, extra_test, body, get_state, set_state, init_vars):
  """Overload of for_stmt that executes a Python for loop."""
  del get_state, set_state

  state = init_vars
  for target in iter_:
    if extra_test is not None and not extra_test(*state):
      break
    state = body(target, *state)
  return state


def _known_len_tf_for_stmt(iter_, extra_test, body, get_state, set_state,
                           init_vars, basic_symbol_names,
                           composite_symbol_names):
  """Overload of for_stmt that iterates over TF entities that admit a length."""
  _disallow_undefs_into_loop(*init_vars)

  n = py_builtins.len_(iter_)
  # TODO(b/117628877): Revisit performance once XLA has the necessary support.
  # Note: using a TensorArray creates an extra copy, but can calculate
  # gradients more efficiently than StridedSlice.
  ta = tensor_array_ops.TensorArray(iter_.dtype, size=n)
  iter_ = ta.unstack(iter_)

  def while_body(iterate_index, *loop_vars):
    """Main loop body."""
    iterate = iter_.read(iterate_index)
    new_vars = body(iterate, *loop_vars)
    _verify_tf_loop_vars(loop_vars, new_vars, basic_symbol_names,
                         composite_symbol_names)

    loop_vars = (iterate_index + 1,)
    if new_vars:
      loop_vars += new_vars

    return loop_vars

  def while_cond(iterate_index, *loop_vars):
    if extra_test is not None:
      return control_flow_ops.cond(
          iterate_index < n, lambda: extra_test(*loop_vars), lambda: False)
    return iterate_index < n

  opts = {}
  # TODO(b/134181679): We do not always set maximum_iterations since that
  # is significantly slower on GPU.
  if control_flow_util.GraphOrParentsInXlaContext(ops.get_default_graph()):
    opts['maximum_iterations'] = n

  results = _tf_while_stmt(
      while_cond,
      while_body,
      get_state,
      set_state,
      (0,) + init_vars,
      None,
      None,
      opts=opts,
  )

  # Note: the iteration index is not returned by the while loop, however
  # if a symbol with the same name exists outside the loop, it will be captured
  # by the loop variables and ultimately updated correctly.
  if isinstance(results, (tuple, list)):
    assert len(results) >= 1  # Has at least the iterate.
    if len(results) > 1:
      results = results[1:]
  else:
    results = ()

  return results


def _tf_range_for_stmt(iter_, extra_test, body, get_state, set_state, init_vars,
                       basic_symbol_names, composite_symbol_names):
  """Overload of for_stmt that iterates over a TF range (and elides it)."""
  _disallow_undefs_into_loop(*init_vars)

  start, limit, delta = iter_.op.inputs

  def while_body(iterate, *loop_vars):
    new_vars = body(iterate, *loop_vars)
    loop_vars = (iterate + delta,)

    if new_vars:
      loop_vars += new_vars

    return loop_vars

  def while_cond(iterate, *loop_vars):
    """Cond function for `tf.while_loop`."""

    def build_main_test():
      """Main iteration condition."""
      # Note(b/138857806): LogicalAnd is slow on GPU so we avoid adding it if
      # `delta` is a compile time constant.
      delta_const = tensor_util.constant_value(delta)
      if delta_const is not None:
        # Support single element arrays.
        delta_const = np.asscalar(delta_const)
        if delta_const >= 0:
          return iterate < limit
        else:
          return iterate > limit
      else:
        return math_ops.logical_or(
            math_ops.logical_and(delta >= 0, iterate < limit),
            math_ops.logical_and(delta < 0, iterate > limit))

    main_test = build_main_test()
    if extra_test is not None:
      return control_flow_ops.cond(
          main_test, lambda: extra_test(*loop_vars), lambda: False)
    return main_test

  # The first loopvar corresponds to the iterate variable which is internal.
  if isinstance(basic_symbol_names, tuple):
    basic_symbol_names = (None,) + basic_symbol_names

  opts = {}
  # TODO(b/134181679): We do not always set maximum_iterations since that
  # is significantly slower on GPU.
  if control_flow_util.GraphOrParentsInXlaContext(ops.get_default_graph()):
    # This specific dtype is required by while_loop.
    opts['maximum_iterations'] = math_ops.cast(
        misc.get_range_len(start, limit, delta), dtypes.int32)

  results = _tf_while_stmt(
      while_cond,
      while_body,
      get_state,
      set_state,
      (start,) + init_vars,
      basic_symbol_names,
      composite_symbol_names,
      opts=opts,
  )

  # Note: the iteration index is not returned by the while loop, however
  # if a symbol with the same name exists outside the loop, it will be captured
  # by the loop variables and ultimately updated correctly.
  if isinstance(results, (tuple, list)):
    assert len(results) >= 1  # Has at least the iterate.
    if len(results) > 1:
      results = results[1:]
  else:
    results = ()

  return results


def _tf_iterator_for_stmt(itr, extra_test, body, get_state, set_state,
                          init_vars, basic_symbol_names,
                          composite_symbol_names):
  """Overload of for_stmt that iterates over TF Iterators. See for_loop."""
  _disallow_undefs_into_loop(*init_vars)

  def while_body_actual(opt_iterate, *loop_vars):
    """Actual main loop body."""
    new_vars = body(opt_iterate.get_value(), *loop_vars)
    _verify_tf_loop_vars(loop_vars, new_vars, basic_symbol_names,
                         composite_symbol_names)
    # TODO(mdan): Fix this inconsistency in the converter.
    if new_vars is None:
      new_vars = ()
    return new_vars

  def while_body(has_next, loop_vars):
    """Main loop body."""
    opt_iterate = iterator_ops.get_next_as_optional(itr)
    has_next = opt_iterate.has_value()

    if not init_vars:
      # cond_v2 requires at least one state tensor in V1.
      dummy_state = (constant_op.constant(()),)
    else:
      dummy_state = ()

    # TODO(mdan): If tf.while_loop supported Optional, this could be avoided.
    new_vars = control_flow_ops.cond(
        has_next,
        lambda: dummy_state + while_body_actual(opt_iterate, *loop_vars),
        lambda: dummy_state + loop_vars,
    )

    if dummy_state:
      new_vars = new_vars[1:]

    return has_next, new_vars

  def while_cond(has_next, loop_vars):
    if extra_test is not None:
      return control_flow_ops.cond(
          has_next, lambda: extra_test(*loop_vars), lambda: False)
    return has_next

  # The first loopvar corresponds to the iterate variable which is internal.
  _, final_vars = _tf_while_stmt(
      while_cond,
      while_body,
      get_state,
      set_state,
      (True, init_vars),
      None,
      None,
      opts=None,
  )
  return final_vars


def _tf_dataset_for_stmt(ds, extra_test, body, get_state, set_state, init_vars,
                         basic_symbol_names, composite_symbol_names):
  """Overload of for_stmt that iterates over TF Datasets."""
  _disallow_undefs_into_loop(*init_vars)

  if extra_test is not None:
    assert init_vars, 'Lowering should always add state.'
    return _dataset_for_stmt_with_extra_test(ds, extra_test, body, get_state,
                                             set_state, init_vars,
                                             basic_symbol_names,
                                             composite_symbol_names)

  return _dataset_for_stmt_no_extra_test(ds, body, get_state, set_state,
                                         init_vars, basic_symbol_names,
                                         composite_symbol_names)


def _dataset_for_stmt_with_extra_test(ds, extra_test, body, get_state,
                                      set_state, init_vars, basic_symbol_names,
                                      composite_symbol_names):
  """Overload of _dataset_for_stmt with early stopping. See for_stmt."""

  # TODO(mdan): Simplify this - following it is extremely difficult.

  def scan_body(aug_vars, iterate):
    """The main loop body wrapper. Only calculates the stop condition."""
    loop_vars, state = aug_vars

    def true_fn():
      set_state(state)
      outputs = body(iterate, *loop_vars)
      _verify_tf_loop_vars(
          loop_vars + state,
          outputs + state,
          basic_symbol_names,
          composite_symbol_names,
          include_shapes=False)
      return outputs, get_state()

    extra_cond = extra_test(*loop_vars)
    new_vars, new_state = control_flow_ops.cond(
        extra_cond, true_fn, lambda: (loop_vars, state))

    scan_outputs = new_vars, new_state, extra_cond
    # Note: new_aug_vars is the actual state of scan; scan_outputs is its output
    # (hence the redundancy).
    # get_state will pull any mutations that body may have made.
    new_aug_vars = new_vars, new_state
    return new_aug_vars, scan_outputs

  def take_while_predicate(unused_loop_vars, unused_state, extra_cond):
    return extra_cond

  def reduce_body(unused_aug_vars, scan_outputs):
    output_aug_vars, output_state, extra_cond = scan_outputs
    del extra_cond
    return output_aug_vars, output_state

  init_state = get_state()
  aug_vars = init_vars, init_state
  ds = ds.apply(scan_ops.scan(aug_vars, scan_body))
  ds = ds.apply(take_while_ops.take_while(take_while_predicate))
  final_aug_vars = ds.reduce(aug_vars, reduce_body)
  final_vars, final_state = final_aug_vars
  set_state(final_state)
  return final_vars


def _dataset_for_stmt_no_extra_test(ds, body, get_state, set_state, init_vars,
                                    basic_symbol_names, composite_symbol_names):
  """Overload of _dataset_for_stmt without early stopping. See for_stmt."""
  init_state = get_state()
  assert isinstance(init_vars, tuple)
  assert isinstance(init_state, tuple)

  # Workaround for Dataset.reduce not allowing empty state tensors - create
  # a dummy state variable that remains unused.
  # TODO(mdan): reduce should allow and match empty structures.
  no_vars = not init_vars
  no_state = not init_state

  if no_vars:
    init_vars = (constant_op.constant(0),)
    if isinstance(basic_symbol_names, tuple):
      basic_symbol_names = (None,) + basic_symbol_names
  if no_state:
    init_state = (constant_op.constant(0),)

  def reduce_body(aug_vars, iterate):
    """The main loop body wrapper."""
    loop_vars, state = aug_vars
    if not no_state:
      set_state(state)

    if no_vars:
      body(iterate)
      new_vars = loop_vars
    else:
      new_vars = body(iterate, *loop_vars)

    if no_state:
      new_state = state
    else:
      new_state = get_state()

    _verify_tf_loop_vars(
        loop_vars + state,
        new_vars + new_state,
        basic_symbol_names,
        composite_symbol_names,
        include_shapes=False)
    return new_vars, new_state

  aug_vars = init_vars, get_state()
  final_vars, final_state = ds.reduce(aug_vars, reduce_body)
  set_state(final_state)

  if no_vars:
    return ()
  return final_vars


def while_stmt(
    test,
    body,
    get_state,
    set_state,
    init_vars,
    basic_symbol_names=None,
    composite_symbol_names=None,
    opts=None,
):
  """Functional form of a while statement.

  The loop operates on a so-called state, which includes all symbols that are
  variant across loop iterations. In what follows we refer to state as either
  a tuple of entities that represent an actual state, or a list of arguments
  of the corresponding types.

  Args:
    test: Callable with the state as arguments, and boolean return type. The
      loop condition.
    body: Callable with the state as arguments, and state as return type. The
      actual loop body.
    get_state: Additional callable which can capture additional state (such as
      the values of composite symbols). This is only useful when staging the
      loop.
    set_state: Additional callable which save values captured by get_state back
      into the Python environment. This is only useful when staging the loop.
    init_vars: Tuple containing the initial state.
    basic_symbol_names: Tuple containing basic loop var names.
    composite_symbol_names: Tuple containing composite loop var names.
    opts: Optional dict of extra loop parameters.

  Returns:
    Tuple containing the final state.
  """

  # Evaluate the initial test once in order to do the dispatch. The evaluation
  # is isolated to minimize unwanted side effects.
  # TODO(mdan): Do a full iteration - some state types might lower to Tensor.
  with func_graph.FuncGraph('tmp').as_default():
    init_test = test(*init_vars)

  # TensorFlow: Multiple evaluations are acceptable in this case, so we're fine
  # with the re-evaluation of `test` that `_tf_while_stmt` will make.
  if tensors.is_dense_tensor(init_test):
    return _tf_while_stmt(test, body, get_state, set_state, init_vars,
                          basic_symbol_names, composite_symbol_names, opts)

  # Normal Python: We already consumed one evaluation of `test`; consistently,
  # unroll one iteration before dispatching to a normal loop.
  # TODO(mdan): Push the "init_test" value via opts into _py_while_stmt?
  if not init_test:
    return init_vars
  init_vars = body(*init_vars)

  return _py_while_stmt(test, body, get_state, set_state, init_vars, opts)


# TODO(kkimlabs): Some callers set basic_symbol_names=None and
# composite_symbol_names=None and call _verify_tf_loop_vars(...) itself.  We can
# remove these arguments once all callers do that.
def _tf_while_stmt(test, body, get_state, set_state, init_vars,
                   basic_symbol_names, composite_symbol_names, opts):
  """Overload of while_stmt that stages a TF while_stmt."""
  _disallow_undefs_into_loop(*init_vars)

  if opts is None:
    opts = {}

  # TODO(mdan): Simplify this.
  loop_vars_slice = slice(len(init_vars))
  state_slice = slice(len(init_vars), None)

  def aug_test(*aug_loop_vars):
    state = aug_loop_vars[state_slice]
    set_state(state)
    return test(*aug_loop_vars[loop_vars_slice])

  def aug_body(*aug_loop_vars):
    state = aug_loop_vars[state_slice]
    set_state(state)
    loop_vars = body(*aug_loop_vars[loop_vars_slice])
    new_state = loop_vars + get_state()
    _verify_tf_loop_vars(aug_loop_vars, new_state, basic_symbol_names,
                         composite_symbol_names)

    return new_state

  # Non-v2 while_loop unpacks the results when there is only one return value.
  # This enforces consistency across versions.
  opts['return_same_structure'] = True

  aug_init_vars = init_vars + get_state()
  final_aug_vars = control_flow_ops.while_loop(aug_test, aug_body,
                                               aug_init_vars, **opts)
  final_state = final_aug_vars[state_slice]
  set_state(final_state)
  return final_aug_vars[loop_vars_slice]


class _PythonLoopChecker(object):
  """Verifies Python loops for TF-specific limits."""

  def __init__(self):
    self.iterations = 0
    self.check_inefficient_unroll = WARN_INEFFICIENT_UNROLL

    # Triggered when we decided to test the op counts.
    self.check_op_count_after_iteration = False

  def _get_ops(self):
    return ops.get_default_graph().get_operations()

  def _check_unroll_limits(self):
    if LIMIT_PYTHON_ITERATIONS and self.iterations > PYTHON_MAX_ITERATIONS:
      raise ValueError('iteration limit exceeded')

  def _stop_checking_inefficient_unroll(self):
    self.check_inefficient_unroll = False
    self.ops_before_iteration = None

  def _verify_ineffcient_unroll(self):
    """Checks for possibly-inefficient creation of ops in a Python loop."""
    assert self.ops_before_iteration is not None
    ops_after_iteration = self._get_ops()
    new_ops = tuple(
        op for op in ops_after_iteration if op not in self.ops_before_iteration)

    if len(new_ops) < INEFFICIENT_UNROLL_MIN_OPS:
      return False

    # TODO(mdan): Add location information.
    ag_logging.warn(
        'TensorFlow ops are being created in a Python loop with large number'
        ' of iterations. This can lead to slow startup. Did you mean to use a'
        ' TensorFlow loop? For example, `while True:` is a Python loop, and'
        ' `while tf.constant(True):` is a TensorFlow loop. The following'
        ' ops were created after iteration %s: %s', self.iterations, new_ops)
    return True

  def before_iteration(self):
    """Called before each iteration in a Python loop."""
    if (self.check_inefficient_unroll and
        self.iterations > INEFFICIENT_UNROLL_MIN_ITERATIONS):
      self.ops_before_iteration = self._get_ops()
      self.check_op_count_after_iteration = True

  def after_iteration(self):
    """Called after each iteration in a Python loop."""
    self.iterations += 1

    self._check_unroll_limits()

    if self.check_inefficient_unroll and self.check_op_count_after_iteration:
      did_warn = self._verify_ineffcient_unroll()
      if did_warn:
        self._stop_checking_inefficient_unroll()  # Only warn once.
      elif self.iterations > INEFFICIENT_UNROLL_MIN_ITERATIONS + 3:
        # Once deciding to check the op counts, only do it for a few iterations.
        self._stop_checking_inefficient_unroll()


def _py_while_stmt(test, body, get_state, set_state, init_vars, opts):
  """Overload of while_stmt that executes a Python while loop."""
  del opts, get_state, set_state

  if __debug__:
    checker = _PythonLoopChecker()

  loop_vars = init_vars
  while test(*loop_vars):

    if __debug__:
      checker.before_iteration()

    loop_vars = body(*loop_vars)

    if __debug__:
      checker.after_iteration()

  return loop_vars


def if_stmt(cond,
            body,
            orelse,
            get_state,
            set_state,
            basic_symbol_names=None,
            composite_symbol_names=None):
  """Functional form of an if statement.

  Args:
    cond: Boolean.
    body: Callable with no arguments, and outputs of the positive (if) branch as
      return type.
    orelse: Callable with no arguments, and outputs of the negative (else)
      branch as return type.
    get_state: Function that returns a tuple containing the values of all
      composite symbols modified within the conditional. This allows access to
      state that branches may mutate through side effects. This function is not
      needed and should not be called when dispatching to code matching Python's
      default semantics. This is useful for checkpointing to avoid unintended
      side-effects when staging requires evaluating all code-paths.
    set_state: Function to set the values of all composite symbols modified
      within the conditional. This is the complement to get_state, used to
      restore checkpointed values. The single argument a tuple containing values
      for each composite symbol that may be modified in a branch of the
      conditional. The is usually the result of a call to get_state.
    basic_symbol_names: Tuple containing basic loop var names.
    composite_symbol_names: Tuple containing composite loop var names.

  Returns:
    Tuple containing the statement outputs.
  """
  # Note: tf.cond doesn't support SparseTensor.
  if tensors.is_dense_tensor(cond):
    return tf_if_stmt(cond, body, orelse, get_state, set_state,
                      basic_symbol_names, composite_symbol_names)
  else:
    return _py_if_stmt(cond, body, orelse)


def tf_if_stmt(cond, body, orelse, get_state, set_state, basic_symbol_names,
               composite_symbol_names):
  """Overload of if_stmt that stages a TF cond."""
  body = _wrap_disallow_undefs_from_cond(body, branch_name='if')
  orelse = _wrap_disallow_undefs_from_cond(orelse, branch_name='else')
  body = _isolate_state(body, get_state, set_state)
  orelse = _isolate_state(orelse, get_state, set_state)

  # `state` currently includes the values of any composite symbols (e.g. `a.b`)
  # composites modified by the loop. `final_vars` includes the values of basic
  # symbols (e.g. `a`) which cannot be passed by reference and must be returned.
  # See _isolate_state.
  # TODO(mdan): We should minimize calls to get/set_state.

  body_branch = 0
  orelse_branch = 1
  result = [None, None]

  def error_checking_body():
    result[body_branch] = body()
    if result[orelse_branch] is not None:
      _verify_tf_cond_vars(result[body_branch], result[orelse_branch],
                           basic_symbol_names, composite_symbol_names)
    return result[body_branch]

  def error_checking_orelse():
    result[orelse_branch] = orelse()
    if result[body_branch] is not None:
      _verify_tf_cond_vars(result[body_branch], result[orelse_branch],
                           basic_symbol_names, composite_symbol_names)
    return result[orelse_branch]

  final_vars, final_state = control_flow_ops.cond(cond, error_checking_body,
                                                  error_checking_orelse)

  set_state(final_state)

  return final_vars


def _isolate_state(func, get_state, set_state):
  """Wraps func to (best-effort) isolate state mutations that func may do.

  The simplest example of state mutation is mutation of variables (via e.g.
  attributes), or modification of globals.

  This allows us to more safely execute this function without worrying about
  side effects when the function wasn't normally expected to execute. For
  example, staging requires that the function is executed ahead of time, and
  we need to ensure its effects are not observed during normal execution.

  Args:
    func: () -> Any
    get_state: () -> Any, returns the current state
    set_state: (Any) -> None, resets the state to the specified values.
      Typically the result of an earlier call to `get_state`.

  Returns:
    Tuple[Any, Any], where the first element is the return value of `func`,
    and the second is the final state values.
  """

  def wrapper():
    init_state = get_state()
    new_vars = func()
    # TODO(mdan): These should be copies, lest set_state might affect them.
    new_state = get_state()
    set_state(init_state)
    return new_vars, new_state

  return wrapper


def _wrap_disallow_undefs_from_cond(func, branch_name):
  """Wraps conditional branch to disallow returning undefined symbols."""

  def wrapper():
    """Calls function and raises an error if undefined symbols are returned."""
    results = func()

    if isinstance(results, tuple):
      results_tuple = results
    else:
      results_tuple = results,
    undefined = tuple(filter(special_values.is_undefined, results_tuple))
    if undefined:
      raise ValueError(
          'The following symbols must also be initialized in the {} branch: {}.'
          ' Alternatively, you may initialize them before the if'
          ' statement.'.format(branch_name,
                               tuple(s.symbol_name for s in undefined)))

    for result in results_tuple:
      if special_values.is_undefined_return(result):
        raise ValueError(
            'A value must also be returned from the {} branch. If a value is '
            'returned from one branch of a conditional a value must be '
            'returned from all branches.'.format(branch_name))

    return results

  return wrapper


def _py_if_stmt(cond, body, orelse):
  """Overload of if_stmt that executes a Python if statement."""
  return body() if cond else orelse()



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""The Autoregressive distribution."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

from tensorflow.python.framework import ops
from tensorflow.python.ops.distributions import distribution as distribution_lib
from tensorflow.python.ops.distributions import util as distribution_util
from tensorflow.python.util import deprecation


class Autoregressive(distribution_lib.Distribution):
  """Autoregressive distributions.

  The Autoregressive distribution enables learning (often) richer multivariate
  distributions by repeatedly applying a [diffeomorphic](
  https://en.wikipedia.org/wiki/Diffeomorphism) transformation (such as
  implemented by `Bijector`s). Regarding terminology,

    "Autoregressive models decompose the joint density as a product of
    conditionals, and model each conditional in turn. Normalizing flows
    transform a base density (e.g. a standard Gaussian) into the target density
    by an invertible transformation with tractable Jacobian." [(Papamakarios et
    al., 2016)][1]

  In other words, the "autoregressive property" is equivalent to the
  decomposition, `p(x) = prod{ p(x[i] | x[0:i]) : i=0, ..., d }`. The provided
  `shift_and_log_scale_fn`, `masked_autoregressive_default_template`, achieves
  this property by zeroing out weights in its `masked_dense` layers.

  Practically speaking the autoregressive property means that there exists a
  permutation of the event coordinates such that each coordinate is a
  diffeomorphic function of only preceding coordinates
  [(van den Oord et al., 2016)][2].

  #### Mathematical Details

  The probability function is

  ```none
  prob(x; fn, n) = fn(x).prob(x)
  ```

  And a sample is generated by

  ```none
  x = fn(...fn(fn(x0).sample()).sample()).sample()
  ```

  where the ellipses (`...`) represent `n-2` composed calls to `fn`, `fn`
  constructs a `tfp.distributions.Distribution`-like instance, and `x0` is a
  fixed initializing `Tensor`.

  #### Examples

  ```python
  import tensorflow_probability as tfp
  tfd = tfp.distributions

  def normal_fn(self, event_size):
    n = event_size * (event_size + 1) / 2
    p = tf.Variable(tfd.Normal(loc=0., scale=1.).sample(n))
    affine = tfd.bijectors.Affine(
        scale_tril=tfd.fill_triangular(0.25 * p))
    def _fn(samples):
      scale = math_ops.exp(affine.forward(samples)).eval()
      return independent_lib.Independent(
          normal_lib.Normal(loc=0., scale=scale, validate_args=True),
          reinterpreted_batch_ndims=1)
    return _fn

  batch_and_event_shape = [3, 2, 4]
  sample0 = array_ops.zeros(batch_and_event_shape)
  ar = autoregressive_lib.Autoregressive(
      self._normal_fn(batch_and_event_shape[-1]), sample0)
  x = ar.sample([6, 5])
  # ==> x.shape = [6, 5, 3, 2, 4]
  prob_x = ar.prob(x)
  # ==> x.shape = [6, 5, 3, 2]

  ```

  #### References

  [1]: George Papamakarios, Theo Pavlakou, and Iain Murray. Masked
       Autoregressive Flow for Density Estimation. In _Neural Information
       Processing Systems_, 2017. https://arxiv.org/abs/1705.07057

  [2]: Aaron van den Oord, Nal Kalchbrenner, Oriol Vinyals, Lasse Espeholt,
       Alex Graves, and Koray Kavukcuoglu. Conditional Image Generation with
       PixelCNN Decoders. In _Neural Information Processing Systems_, 2016.
       https://arxiv.org/abs/1606.05328
  """

  @deprecation.deprecated(
      "2018-10-01",
      "The TensorFlow Distributions library has moved to "
      "TensorFlow Probability "
      "(https://github.com/tensorflow/probability). You "
      "should update all references to use `tfp.distributions` "
      "instead of `tf.contrib.distributions`.",
      warn_once=True)
  def __init__(self,
               distribution_fn,
               sample0=None,
               num_steps=None,
               validate_args=False,
               allow_nan_stats=True,
               name="Autoregressive"):
    """Construct an `Autoregressive` distribution.

    Args:
      distribution_fn: Python `callable` which constructs a
        `tfp.distributions.Distribution`-like instance from a `Tensor` (e.g.,
        `sample0`). The function must respect the "autoregressive property",
        i.e., there exists a permutation of event such that each coordinate is a
        diffeomorphic function of on preceding coordinates.
      sample0: Initial input to `distribution_fn`; used to
        build the distribution in `__init__` which in turn specifies this
        distribution's properties, e.g., `event_shape`, `batch_shape`, `dtype`.
        If unspecified, then `distribution_fn` should be default constructable.
      num_steps: Number of times `distribution_fn` is composed from samples,
        e.g., `num_steps=2` implies
        `distribution_fn(distribution_fn(sample0).sample(n)).sample()`.
      validate_args: Python `bool`.  Whether to validate input with asserts.
        If `validate_args` is `False`, and the inputs are invalid,
        correct behavior is not guaranteed.
      allow_nan_stats: Python `bool`, default `True`. When `True`, statistics
        (e.g., mean, mode, variance) use the value "`NaN`" to indicate the
        result is undefined. When `False`, an exception is raised if one or
        more of the statistic's batch members are undefined.
      name: Python `str` name prefixed to Ops created by this class.
        Default value: "Autoregressive".

    Raises:
      ValueError: if `num_steps` and
        `distribution_fn(sample0).event_shape.num_elements()` are both `None`.
      ValueError: if `num_steps < 1`.
    """
    parameters = dict(locals())
    with ops.name_scope(name) as name:
      self._distribution_fn = distribution_fn
      self._sample0 = sample0
      self._distribution0 = (distribution_fn() if sample0 is None
                             else distribution_fn(sample0))
      if num_steps is None:
        num_steps = self._distribution0.event_shape.num_elements()
        if num_steps is None:
          raise ValueError("distribution_fn must generate a distribution "
                           "with fully known `event_shape`.")
      if num_steps < 1:
        raise ValueError("num_steps ({}) must be at least 1.".format(num_steps))
      self._num_steps = num_steps
    super(Autoregressive, self).__init__(
        dtype=self._distribution0.dtype,
        reparameterization_type=self._distribution0.reparameterization_type,
        validate_args=validate_args,
        allow_nan_stats=allow_nan_stats,
        parameters=parameters,
        graph_parents=self._distribution0._graph_parents,  # pylint: disable=protected-access
        name=name)

  @property
  def distribution_fn(self):
    return self._distribution_fn

  @property
  def sample0(self):
    return self._sample0

  @property
  def num_steps(self):
    return self._num_steps

  @property
  def distribution0(self):
    return self._distribution0

  def _batch_shape(self):
    return self.distribution0.batch_shape

  def _batch_shape_tensor(self):
    return self.distribution0.batch_shape_tensor()

  def _event_shape(self):
    return self.distribution0.event_shape

  def _event_shape_tensor(self):
    return self.distribution0.event_shape_tensor()

  def _sample_n(self, n, seed=None):
    if seed is None:
      seed = distribution_util.gen_new_seed(
          seed=np.random.randint(2**32 - 1),
          salt="autoregressive")
    samples = self.distribution0.sample(n, seed=seed)
    for _ in range(self._num_steps):
      samples = self.distribution_fn(samples).sample(seed=seed)
    return samples

  def _log_prob(self, value):
    return self.distribution_fn(value).log_prob(value)

  def _prob(self, value):
    return self.distribution_fn(value).prob(value)



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Functions used to extract and analyze stacks.  Faster than Python libs."""
# pylint: disable=g-bad-name
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import inspect
import linecache
import threading

import six

# TODO(b/138203821): change to from ...util import ... once the bug is fixed.
from tensorflow.python import _tf_stack

# Generally such lookups should be done using `threading.local()`. See
# https://blogs.gnome.org/jamesh/2008/06/11/tls-python/ for a detailed
# explanation of why. However the transform stacks are expected to be empty
# when a thread is joined, so reusing the key does not introduce a correctness
# issue. Moreover, get_ident is faster than storing and retrieving a unique
# key in a thread local store.
if six.PY3:
  _get_thread_key = threading.get_ident
else:
  import thread  # pylint: disable=g-import-not-at-top
  _get_thread_key = thread.get_ident


# Names for indices into TF traceback tuples.
TB_FILENAME = 0
TB_LINENO = 1
TB_FUNCNAME = 2
TB_CODEDICT = 3  # Dictionary of Python interpreter state.


_source_mapper_stacks = collections.defaultdict(list)
_source_filter_stacks = collections.defaultdict(list)


class StackTraceTransform(object):
  """Base class for stack trace transformation functions."""

  _stack_dict = None  # Subclasses should override
  _thread_key = None

  def __enter__(self):
    self.reset()

    # Any given instance is assumed to be used by a single thread, which reduces
    # expensive thread local lookups.
    if self._thread_key is None:
      self._thread_key = _get_thread_key()
    else:
      assert self._thread_key == _get_thread_key(), 'Shared across threads?'

    stack = self._stack_dict[self._thread_key]
    if stack:
      self.parent = stack[-1]
    else:
      self.parent = None
    stack.append(self)
    return self

  def __exit__(self, unused_type, unused_value, unused_traceback):
    top = self._stack_dict[self._thread_key].pop()
    assert top is self, 'Concurrent access?'

  def reset(self):
    pass


class StackTraceMapper(StackTraceTransform):
  """Allows remapping traceback information to different source code."""
  _stack_dict = _source_mapper_stacks

  def reset(self):
    self._effective_source_map = None

  def get_effective_source_map(self):
    """Returns a map (filename, lineno) -> (filename, lineno, function_name)."""
    raise NotImplementedError('subclasses need to override this')


class StackTraceFilter(StackTraceTransform):
  """Allows filtering traceback information by removing superfluous frames."""
  _stack_dict = _source_filter_stacks

  def reset(self):
    self._filtered_filenames = None

  def get_filtered_filenames(self):
    raise NotImplementedError('subclasses need to override this')


class CurrentModuleFilter(StackTraceFilter):
  """Filters stack frames from the module where this is used (best effort)."""

  def __init__(self):
    filter_filename = None
    outer_f = None
    f = inspect.currentframe()
    try:
      if f is not None:
        # The current frame is __init__. The first outer frame should be the
        # caller.
        outer_f = f.f_back
        if outer_f is not None:
          filter_filename = inspect.getsourcefile(outer_f)
      self._filename = filter_filename
    finally:
      # Avoid reference cycles, see:
      # https://docs.python.org/3.7/library/inspect.html#the-interpreter-stack
      del f
      del outer_f

  def get_filtered_filenames(self):
    if self._filtered_filenames is None:
      self._filtered_filenames = frozenset((self._filename,))
      if self.parent is not None:
        self._filtered_filenames |= self.parent.get_filtered_filenames()
    return self._filtered_filenames


def extract_stack(limit=-1):
  """A lightweight, extensible re-implementation of traceback.extract_stack.

  NOTE(mrry): traceback.extract_stack eagerly retrieves the line of code for
      each stack frame using linecache, which results in an abundance of stat()
      calls. This implementation does not retrieve the code, and any consumer
      should apply _convert_stack to the result to obtain a traceback that can
      be formatted etc. using traceback methods.

  Args:
    limit: A limit on the number of frames to return.

  Returns:
    A sequence of StackFrame objects
        (filename, lineno, name, globals, func_start_lineno)
    corresponding to the call stack of the current thread.  The returned
    tuples have the innermost stack frame at the end, unlike the Python
    inspect module's stack() function.
  """
  # N.B ExtractStack in tf_stack.cc will drop this frame prior to
  # traversing the stack.
  thread_key = _get_thread_key()
  return _tf_stack.extract_stack(
      limit,
      _source_mapper_stacks[thread_key],
      _source_filter_stacks[thread_key])

StackFrame = _tf_stack.StackFrame


def convert_stack(stack, include_func_start_lineno=False):
  """Converts a stack extracted using extract_stack() to a traceback stack.

  Args:
    stack: A list of n 5-tuples,
      (filename, lineno, name, frame_globals, func_start_lineno).
    include_func_start_lineno: True if function start line number should be
      included as the 5th entry in return tuples.

  Returns:
    A tuple of n 4-tuples or 5-tuples
    (filename, lineno, name, code, [optional: func_start_lineno]), where the
    code tuple element is calculated from the corresponding elements of the
    input tuple.
  """
  def _tuple_generator():  # pylint: disable=missing-docstring
    for frame in stack:
      filename = frame.filename
      lineno = frame.lineno
      linecache.checkcache(filename)
      line = linecache.getline(filename, lineno, frame.globals)
      if line:
        line = line.strip()
      else:
        line = None
      if include_func_start_lineno:
        yield (filename, lineno, frame.name, line, frame.func_start_lineno)
      else:
        yield (filename, lineno, frame.name, line)

  return tuple(_tuple_generator())



# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Reversible residual network compatible with eager execution.

Code for main model.

Reference [The Reversible Residual Network: Backpropagation
Without Storing Activations](https://arxiv.org/pdf/1707.04585.pdf)
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf
from tensorflow.contrib.eager.python.examples.revnet import blocks


class RevNet(tf.keras.Model):
  """RevNet that depends on all the blocks."""

  def __init__(self, config):
    """Initialize RevNet with building blocks.

    Args:
      config: tf.contrib.training.HParams object; specifies hyperparameters
    """
    super(RevNet, self).__init__(dtype=config.dtype)
    self.axis = 1 if config.data_format == "channels_first" else 3
    self.config = config

    self._init_block = blocks.InitBlock(config=self.config)
    self._final_block = blocks.FinalBlock(config=self.config)
    self._block_list = self._construct_intermediate_blocks()
    self._moving_average_variables = []

  def _construct_intermediate_blocks(self):
    # Precompute input shape after initial block
    stride = self.config.init_stride
    if self.config.init_max_pool:
      stride *= 2
    if self.config.data_format == "channels_first":
      w, h = self.config.input_shape[1], self.config.input_shape[2]
      input_shape = (self.config.init_filters, w // stride, h // stride)
    else:
      w, h = self.config.input_shape[0], self.config.input_shape[1]
      input_shape = (w // stride, h // stride, self.config.init_filters)

    # Aggregate intermediate blocks
    block_list = tf.contrib.checkpoint.List()
    for i in range(self.config.n_rev_blocks):
      # RevBlock configurations
      n_res = self.config.n_res[i]
      filters = self.config.filters[i]
      if filters % 2 != 0:
        raise ValueError("Number of output filters must be even to ensure"
                         "correct partitioning of channels")
      stride = self.config.strides[i]
      strides = (self.config.strides[i], self.config.strides[i])

      # Add block
      rev_block = blocks.RevBlock(
          n_res,
          filters,
          strides,
          input_shape,
          batch_norm_first=(i != 0),  # Only skip on first block
          data_format=self.config.data_format,
          bottleneck=self.config.bottleneck,
          fused=self.config.fused,
          dtype=self.config.dtype)
      block_list.append(rev_block)

      # Precompute input shape for the next block
      if self.config.data_format == "channels_first":
        w, h = input_shape[1], input_shape[2]
        input_shape = (filters, w // stride, h // stride)
      else:
        w, h = input_shape[0], input_shape[1]
        input_shape = (w // stride, h // stride, filters)

    return block_list

  def call(self, inputs, training=True):
    """Forward pass."""

    saved_hidden = None
    if training:
      saved_hidden = [inputs]

    h = self._init_block(inputs, training=training)
    if training:
      saved_hidden.append(h)

    for block in self._block_list:
      h = block(h, training=training)
      if training:
        saved_hidden.append(h)

    logits = self._final_block(h, training=training)

    return (logits, saved_hidden) if training else (logits, None)

  def compute_loss(self, logits, labels):
    """Compute cross entropy loss."""

    if self.config.dtype == tf.float32 or self.config.dtype == tf.float16:
      cross_ent = tf.nn.sparse_softmax_cross_entropy_with_logits(
          logits=logits, labels=labels)
    else:
      # `sparse_softmax_cross_entropy_with_logits` does not have a GPU kernel
      # for float64, int32 pairs
      labels = tf.one_hot(
          labels, depth=self.config.n_classes, axis=1, dtype=self.config.dtype)
      cross_ent = tf.nn.softmax_cross_entropy_with_logits(
          logits=logits, labels=labels)

    return tf.reduce_mean(cross_ent)

  def compute_gradients(self, saved_hidden, labels, training=True, l2_reg=True):
    """Manually computes gradients.

    This method silently updates the running averages of batch normalization.

    Args:
      saved_hidden: List of hidden states Tensors
      labels: One-hot labels for classification
      training: Use the mini-batch stats in batch norm if set to True
      l2_reg: Apply l2 regularization

    Returns:
      A tuple with the first entry being a list of all gradients and the second
      being the loss
    """

    def _defunable_pop(l):
      """Functional style list pop that works with `tfe.defun`."""
      t, l = l[-1], l[:-1]
      return t, l

    # Backprop through last block
    x = saved_hidden[-1]
    with tf.GradientTape() as tape:
      tape.watch(x)
      logits = self._final_block(x, training=training)
      loss = self.compute_loss(logits, labels)
    grads_combined = tape.gradient(loss,
                                   [x] + self._final_block.trainable_variables)
    dy, final_grads = grads_combined[0], grads_combined[1:]

    # Backprop through intermediate blocks
    intermediate_grads = []
    for block in reversed(self._block_list):
      y, saved_hidden = _defunable_pop(saved_hidden)
      x = saved_hidden[-1]
      dy, grads = block.backward_grads(x, y, dy, training=training)
      intermediate_grads = grads + intermediate_grads

    # Backprop through first block
    _, saved_hidden = _defunable_pop(saved_hidden)
    x, saved_hidden = _defunable_pop(saved_hidden)
    assert not saved_hidden
    with tf.GradientTape() as tape:
      y = self._init_block(x, training=training)
    init_grads = tape.gradient(
        y, self._init_block.trainable_variables, output_gradients=dy)

    # Ordering match up with `model.trainable_variables`
    grads_all = init_grads + final_grads + intermediate_grads
    if l2_reg:
      grads_all = self._apply_weight_decay(grads_all)

    return grads_all, loss

  def _apply_weight_decay(self, grads):
    """Update gradients to reflect weight decay."""
    return [
        g + self.config.weight_decay * v if v.name.endswith("kernel:0") else g
        for g, v in zip(grads, self.trainable_variables)
    ]

  def get_moving_stats(self):
    """Get moving averages of batch normalization."""
    device = "/gpu:0" if tf.test.is_gpu_available() else "/cpu:0"
    with tf.device(device):
      return [v.read_value() for v in self.moving_average_variables]

  def restore_moving_stats(self, values):
    """Restore moving averages of batch normalization."""
    device = "/gpu:0" if tf.test.is_gpu_available() else "/cpu:0"
    with tf.device(device):
      for var_, val in zip(self.moving_average_variables, values):
        var_.assign(val)

  @property
  def moving_average_variables(self):
    """Get all variables that are batch norm moving averages."""

    def _is_moving_avg(v):
      n = v.name
      return n.endswith("moving_mean:0") or n.endswith("moving_variance:0")

    if not self._moving_average_variables:
      self._moving_average_variables = filter(_is_moving_avg, self.variables)

    return self._moving_average_variables
